import asyncio
import json

import asyncws

from . import PullBase
from ...utils import make_list, auto_str_ignore, include_or_exclude, wildcards_to_regex


@auto_str_ignore(['token', '_websocket', '_loop', '_include_regex', '_exclude_regex'])
class State(PullBase):
    __prefix__ = 'hass'

    def __init__(self, url, token, include=None, exclude=None, **kwargs):
        super().__init__(**kwargs)
        self.url = self._sanitize_url(url)
        self.token = str(token)
        self.include = make_list(include)
        self.exclude = make_list(exclude)
        self._websocket = None
        self._loop = None

        self._include_regex = self.include and wildcards_to_regex(self.include)
        self._exclude_regex = self.exclude and wildcards_to_regex(self.exclude)

    @staticmethod
    def _sanitize_url(url):
        res = str(url).replace('https://', 'wss://').replace('http://', 'ws://')
        if res.endswith('/'):
            return res[:-1]  # Remove last /
        return res

    def _layout_message(self, message):
        def _layout_state(state):
            from copy import copy
            cp = copy(state)
            cp.pop('entity_id', None)
            cp.pop('context', None)
            return cp

        state_data = message.get('event', {}).get('data', {})
        entity_id = state_data.get('entity_id')
        old_state = _layout_state(state_data.get('old_state', {}))
        new_state = _layout_state(state_data.get('new_state', {}))
        return {'entity_id': entity_id, 'old_state': old_state, 'new_state': new_state}

    def _emit(self, message):
        payload = self._layout_message(message)
        if include_or_exclude(payload['entity_id'], self._include_regex, self._exclude_regex):
            self.on_payload(self, payload)

    async def _receive_states(self):
        self._websocket = await asyncws.connect('{self.url}/api/websocket'.format(**locals()))

        while True:
            message = await self._websocket.recv()
            if message is None:
                break
            message = json.loads(message)
            if message.get('type', '') == 'auth_required':
                hass_version = message.get('ha_version')
                self.logger.info("[{self.name}] Connected to Home Assistant {hass_version} websocket @ {self.url}"
                                 .format(**locals()))
                self.logger.debug("[{self.name}] Authentication is required. Providing token".format(**locals()))
                await self._websocket.send(json.dumps({
                    'type': 'auth',
                    'access_token': self.token
                }))
            elif message.get('type', '') == 'auth_ok':
                self.logger.info("[{self.name}] Authentication is valid".format(**locals()))
                await self._websocket.send(json.dumps(
                    {'id': 1, 'type': 'subscribe_events', 'event_type': 'state_changed'}
                ))
            elif message.get('type', '') == 'auth_invalid':
                self.logger.info("[{self.name}] Authentication is invalid. Aborting...".format(**locals()))
                await self._websocket.close()
            elif message.get('type', '') == 'result':
                pass
            elif message.get('type', '') == 'event':
                self._emit(message)
            else:
                self.logger.warning("[{self.name}] Got unexpected message '{message}'".format(**locals()))

    def stop(self):
        super().stop()
        if self._loop:
            asyncio.run_coroutine_threadsafe(self._websocket.close(), self._loop)

    def pull(self):
        try:
            self._loop = asyncio.get_event_loop()
            is_new = False
        except RuntimeError:
            self._loop = asyncio.new_event_loop()
            is_new = True
        try:
            self._loop.run_until_complete(self._receive_states())
        finally:
            if is_new:
                self._loop.close()
            self._loop = None
            self._websocket = None



from . import PullBase


class Subscribe(PullBase):
    """
    Pulls messages from the specified topic from the given mosquitto mqtt broker (identified by host and port).

    All arguments (`host`, `port` and `topic`) can be automatically injected via environment variables.
    * MQTT_HOST
    * MQTT_PORT
    * MQTT_TOPIC

    Args:
        host (str): The host where the mosquitto broker is running.
        port (int): The port where the mosquitto broker is listening (default is 1883).
        topic (str): The topic to publish to.

    Returns:
        The callback `on_payload` will offer a payload that is a dictionary of the received event.
        Example ->
        {
            'topic': 'test/device/device1',
            'levels': ['test', 'device', 'device1']
            'payload': 'The actual event message'
        }

    Example configuration:

        name: mqtt
        pull:
          plugin: pnp.plugins.pull.mqtt.MQTTPull
          args:
            host: localhost
            port: 1883
            topic: test/#
        push:
          plugin: pnp.plugins.push.simple.Echo

    """
    __prefix__ = 'mqtt'

    def __init__(self, host, topic, port=1883, user=None, password=None, **kwargs):
        super().__init__(**kwargs)
        self.host = str(host)
        self.topic = str(topic)
        self.port = int(port)
        self.user = user and str(user)
        self.password = password and str(password)
        self._client = None

    def on_connect(self, client, userdata, flags, rc):
        # Subscribing in on_connect() means that if we lose the connection and
        # reconnect then subscriptions will be renewed.
        if rc == 0:
            client.subscribe(self.topic)
            self.logger.info("[{self.name}] Connected with result code '{rc}' "
                             "to {self.topic} @ {self.host}:{self.port}".format(**locals()))
        else:
            self.logger.error("[{self.name}] Bad connection with result code '{rc}' "
                              "to {self.topic} @ {self.host}:{self.port}".format(**locals()))

    def on_disconnect(self, client, userdata, rc):
        if rc != 0:
            self.logger.warning("[{self.name}] Unexpected mqtt disconnect with result code '{rc}'. "
                                "Will automatically reconnect.".format(**locals()))

    def on_message(self, client, obj, msg):
        self.logger.debug("[{self.name}] Got message from broker on topic '{self.topic}'. "
                          "Payload='{msg.payload}'".format(**locals()))

        # This one is a envelope with data
        self.notify(dict(
            topic=str(msg.topic),
            levels=[level for level in msg.topic.split('/')],
            payload=msg.payload.decode('utf-8')
        ))

    def stop(self):
        super().stop()
        if self._client:
            self._client.disconnect()

    def pull(self):
        import paho.mqtt.client as paho

        self._client = paho.Client()
        if self.user:
            self._client.username_pw_set(self.user, self.password)
        self._client.on_connect = self.on_connect
        self._client.on_message = self.on_message
        self._client.on_disconnect = self.on_disconnect

        self._client.connect(self.host, self.port, 60)
        self._client.loop_forever(retry_first_connection=True)


# For backwards compat
MQTTPull = Subscribe



from .models import UDFModel

from .utils import Singleton, safe_eval, FallbackBox, EvaluationError
from .validator import Validator


class PayloadSelector(Singleton):
    """
    Examples:

        >>> dut = PayloadSelector.instance

        >>> dut.eval_selector('payload.k1', dict(k1='k1', k2='k2'))
        'k1'
        >>> dut.eval_selector('suppress_me if int(payload) == 0 else int(payload)', 1)
        1
        >>> dut.eval_selector('suppress_me if int(payload) == 0 else int(payload)', 0) is dut.SuppressLiteral
        True
        >>> dut.eval_selector('on_off(data)', True)
        'on'

        >>> dut.instance.register_custom_global('on_off_custom', lambda x: 'on' if x else 'off')
        >>> dut.eval_selector('on_off_custom(data)', True)
        'on'
        >>> dut.eval_selector('on_off_custom(data)', False)
        'off'

        >>> payload = {'foo': {'bar': 1, 'baz': 2}}
        >>> dut.eval_selector({
        ...     "lambda payload: payload['foo']['bar']": 'baz',
        ...     "foo": "lambda payload: payload['foo']['baz']",
        ...     "lambda payload: payload['foo']['baz']": "lambda payload: payload['foo']['bar']"
        ... }, payload=payload) == {1: 'baz', 'foo': 2, 2: 1}
        True

        >>> dut.eval_selector([
        ...     "foo",
        ...     "lambda payload: payload['foo']['bar']",
        ...     "lambda payload: payload['foo']['baz']",
        ...     {"lambda payload: payload['foo']['bar']": "lambda payload: payload['foo']['baz']"},
        ...     ["foo", "bar", "lambda payload: payload['foo']['bar']"]
        ... ], payload=payload)
        ['foo', 1, 2, {1: 2}, ['foo', 'bar', 1]]

        >>> dut.eval_selector("payload", payload=payload) == {'foo': {'bar': 1, 'baz': 2}}
        True
        >>> dut.eval_selector({"payload": "payload"}, payload=payload)
        {'payload': 'payload'}
        >>> (dut.eval_selector({"payload": "lambda payload: payload"}, payload=payload)
        ...     == {'payload': {'foo': {'bar': 1, 'baz': 2}}})
        True

        >>> dut.eval_selector("this one is not known", payload=payload)
        Traceback (most recent call last):
        ...
        pnp.utils.EvaluationError: Failed to evaluate 'this one is not known'
        >>> dut.eval_selector(["lambda payload: known this is not"], payload=payload)
        Traceback (most recent call last):
        ...
        pnp.utils.EvaluationError: Your lambda is errorneous: 'lambda payload: known this is not'
        >>> dut.eval_selector(["lambda payload: known"], payload=payload)
        Traceback (most recent call last):
        ...
        pnp.utils.EvaluationError: Error when running the selector lambda: 'lambda payload: known'
        >>> dut.eval_selector({'str': 'str'}, payload=payload)  # Callables but no lambdas
        {'str': 'str'}
    """

    def __init__(self):
        self._suppress_literal = object()
        self._custom = {}
        self._register_globals()

    @property
    def SuppressLiteral(self):
        return self._suppress_literal

    @property
    def suppress_aliases(self):
        uppers = ["SUPPRESS", "SUPPRESSME", "SUPPRESSPUSH", "SUPPRESS_ME", "SUPPRESS_PUSH"]
        return uppers + [item.lower() for item in uppers]

    def _register_globals(self):
        self._custom["abs"] = abs
        self._custom["bool"] = bool
        self._custom["dict"] = dict
        self._custom["float"] = float
        self._custom["getattr"] = getattr
        self._custom["hasattr"] = hasattr
        self._custom["hash"] = hash
        self._custom["int"] = int
        self._custom["isinstance"] = isinstance
        self._custom["len"] = len
        self._custom["list"] = list
        self._custom["max"] = max
        self._custom["min"] = min
        self._custom["ord"] = ord
        self._custom["pow"] = pow
        self._custom["reversed"] = reversed
        self._custom["round"] = round
        self._custom["sorted"] = sorted
        self._custom["str"] = str
        self._custom["zip"] = zip

        from base64 import b64encode, b64decode
        self._custom["b64encode"] = b64encode
        self._custom["b64decode"] = b64decode
        import os
        self._custom["basename"] = os.path.basename
        from .utils import on_off
        self._custom["on_off"] = on_off

    def register_custom_global(self, name, fun):
        """
        Register a custom function that should be available in the context of the selector.

        Args:
            name: Name of the function to register.
            fun: The actual function to register.

        Returns:
            None.
        """
        Validator.is_function(fun=fun)
        Validator.is_instance(str, name=name)
        if name not in self._custom:  # Overriding a already existing custom will silently fail!
            self._custom[name] = fun

    def register_udfs(self, udfs):
        Validator.is_iterable_but_no_str(udfs=udfs)
        for i, udf in enumerate(udfs):
            Validator.is_instance(UDFModel, **{'udfs.item_{i}'.format(**locals()): udf})
            self.register_custom_global(udf.name, udf.callable)

    def _eval_wrapper(self, selector, payload):
        suppress_kwargs = {alias: self.SuppressLiteral for alias in self.suppress_aliases}
        return safe_eval(
            source=selector,
            payload=payload,
            data=payload,
            **suppress_kwargs,
            **self._custom
        )

    @staticmethod
    def _isalambda(v):
        LAMBDA = lambda: 0  # pylint: disable=E731
        return isinstance(v, type(LAMBDA)) and v.__name__ == LAMBDA.__name__

    def _e(self, snippet, payload, is_key=False):
        # There might be nested dicts or lists inside the complex structure... Eval them recursively
        if isinstance(snippet, dict):
            return {self._e(k, payload, True): self._e(v, payload) for k, v in snippet.items()}
        if isinstance(snippet, list):
            return [self._e(i, payload) for i in snippet]

        # Test if the snippet constructs a lambda -> assumes to be callable code with payload argument
        try:
            possible_fun = self._eval_wrapper(snippet, payload)
        except EvaluationError:
            if str(snippet).startswith('lambda'):
                raise EvaluationError("Your lambda is errorneous: '{snippet}'".format(**locals()))
            return snippet  # Evaluation failed -> assume that is string literal (dict key/value or list item)

        if not self._isalambda(possible_fun):
            return snippet  # Cannot be executed. So assume it is not a selector expression

        # It is a callable. Lets modify globals and call it :-)
        suppress_kwargs = {alias: self.SuppressLiteral for alias in self.suppress_aliases}
        customs = self._custom
        for k, v in {**suppress_kwargs, **customs}.items():
            possible_fun.__globals__[k] = v
        try:
            return possible_fun(payload)
        except:
            raise EvaluationError("Error when running the selector lambda: '{snippet}'".format(**locals()))

    def eval_selector(self, selector, payload):
        # Wrap payload inside a Box -> this makes dot accessable dictionaries possible
        # We create a dictionary cause payload might not be an actual dictionary.
        # This way wrapping with Box will always work ;-)
        payload = FallbackBox({'base': payload}).base
        if selector is None:
            return payload

        if isinstance(selector, (list, dict)):
            return self._e(selector, payload)  # Complex. Need additional magic

        # No complex structure. We assume that is an expression and we try to evaluate it
        return self._eval_wrapper(selector, payload)



import json

import requests

from ...utils import try_parse_bool
from ...validator import Validator
from ... import utils
from . import PushBase, PushExecutionError


class Call(PushBase):

    def __init__(self, url, method='GET', fail_on_error=False, provide_response=False, **kwargs):
        super().__init__(**kwargs)
        self.url = self._parse_url(url)
        self.method = self._parse_method(method)
        self.fail_on_error = self._parse_fail_on_error(fail_on_error)
        self.provide_response = bool(provide_response)

    def _parse_url(self, val):
        return str(val)

    def _parse_method(self, val):
        val = str(val).upper()
        Validator.one_of(utils.HTTP_METHODS, method=val)
        return val

    def _parse_fail_on_error(self, val):
        return try_parse_bool(val)

    def push(self, payload):
        envelope, real_payload = self.envelope_payload(payload)

        url = self._parse_envelope_value('url', envelope)
        method = self._parse_envelope_value('method', envelope)
        fail_on_error = self._parse_envelope_value('fail_on_error', envelope)

        if isinstance(real_payload, (dict, list, tuple)):
            try:
                real_payload = json.dumps(real_payload)
            except:
                pass
        resp = requests.request(method, url, data=str(real_payload))
        if fail_on_error and not (200 <= resp.status_code <= 299):
            raise PushExecutionError("{method} of '{url}' failed with "
                                     "status code = '{resp.status_code}'".format(**locals()))
        if not self.provide_response:
            return payload

        try:
            return dict(status_code=resp.status_code, is_json=True, data=resp.json())
        except ValueError:
            # No valid json, try text
            return dict(status_code=resp.status_code, is_json=False, data=resp.text)



import json

import requests

from . import PullBase
from .. import load_optional_module
from ...utils import auto_str_ignore, make_list, HTTP_METHODS
from ...validator import Validator


@auto_str_ignore(ignore_list=['server'])
class Server(PullBase):
    """
    Listens on the specified `port` for requests to any endpoint.
    Any data passed to the endpoint will be tried to be parsed to a dictionary (json). If this is not possible
    the data will be passed as is. See sections `Returns` for specific payload and examples.

    Remark: You will not able to make requests to the endpoint DELETE `/_shutdown` because it is used internally.

    Args:
        port (int): The port the rest server should listen to for requests.

    Returns:
        The callback `on_payload` will offer a payload that is a dictionary of the request made.

        Examples:
            curl -X GET 'http://localhost:5000/resource/endpoint?foo=bar&bar=baz' --data '{"baz": "bar"}'
            {
                'endpoint': 'resource/endpoint,
                'method': 'GET',
                'query': {'foo': 'bar', 'bar': 'baz'},
                'data': {'baz': 'bar'},
                'is_json': True
            }
            curl -X GET 'http://localhost:5000/resource/endpoint' --data 'no json obviously'
            {
                'endpoint': 'resource/endpoint,
                'method': 'GET',
                'query': {},
                'data': b'no json obviously',
                'is_json': False
            }

    Example configuration:

        name: rest
        pull:
          plugin: pnp.plugins.pull.rest.RestServer
          args:
            port: 5000
            allowed_methods: [GET, POST]
        push:
          plugin: pnp.plugins.push.simple.Echo

    """
    __prefix__ = "rest"

    EXTRA = 'http-server'
    SERVER_IMPL = ['gevent', 'flask']

    def __init__(self, port=5000, allowed_methods='GET', server_impl='gevent', **kwargs):
        super().__init__(**kwargs)
        self.port = int(port)
        self.allowed_methods = [m.upper() for m in make_list(allowed_methods)]
        Validator.subset_of(HTTP_METHODS, allowed_methods=self.allowed_methods)

        self.server_impl = server_impl
        Validator.one_of(self.SERVER_IMPL, server_impl=self.server_impl)

        # Instance of the actual http server (wsgi)
        self.server = None

    def pull(self):
        app = self._create_app()
        if self.server_impl == 'flask':
            # Running the flask app with flask development server (do not use for production!)
            app.run(host='0.0.0.0', port=self.port, threaded=True)
            self.server = app
        elif self.server_impl == 'gevent':
            WSGIServer = load_optional_module('gevent.pywsgi', self.EXTRA).WSGIServer
            self.server = WSGIServer(('', self.port), app)
            self.server.serve_forever()

    def stop(self):
        super().stop()
        if self.server_impl == 'flask':
            requests.delete('http://localhost:{port}/_shutdown'.format(port=str(self.port)))
        elif self.server_impl == 'gevent':
            if self.server:
                self.server.stop()

    def _create_app(self):
        that = self
        Flask = load_optional_module('flask', self.EXTRA).Flask
        app = Flask(__name__)

        if self.server_impl == 'flask':
            # We need to register a shutdown endpoint, to end the serving if using the flask development server
            @app.route('/_shutdown', methods=['DELETE'])
            def shutdown():
                from flask import request
                func = request.environ.get('werkzeug.server.shutdown')
                if func is None:
                    raise RuntimeError('Not running with the Werkzeug Server')  # pragma: no cover
                func()
                return json.dumps({'success': True}), 200, {'ContentType': 'application/json'}

        @app.route('/', defaults={'path': '/'}, methods=self.allowed_methods)
        @app.route('/<path:path>', methods=self.allowed_methods)
        def catch_all(path):
            from flask import request
            data = request.get_json(force=True, silent=True)
            if data is None:  # No valid json in request body > fallback to data
                data = request.data if request.data != b'' else None

            payload = dict(
                endpoint=path,
                levels=["/"] if path == "/" else path.split('/'),
                method=request.method,
                query=self._flatten_query_args(dict(request.args)),
                data=data,
                is_json=isinstance(data, dict),
                url=request.url,
                full_path=request.full_path,
                path=request.path
            )
            that.notify(payload)

            return json.dumps({'success': True}), 200, {'ContentType': 'application/json'}

        return app

    @staticmethod
    def _flatten_query_args(args):
        """Iterates through query args and transforms any one-element lists to single items.

        Examples:

            >>> Server._flatten_query_args({'key': 'value'})  # Return as is
            {'key': 'value'}
            >>> Server._flatten_query_args({'key': ['value']})  # One item list -> flatten
            {'key': 'value'}
            >>> Server._flatten_query_args({'key': ['value1', 'value2']})  # multiple items list -> no flatten
            {'key': ['value1', 'value2']}
            >>> Server._flatten_query_args({'key': ['']})  # Empty string -> None
            {'key': None}
            >>> Server._flatten_query_args({'key': []})  # Empty list -> None
            {'key': None}
            >>> Server._flatten_query_args({'key': ['', '']})  # Multiple Empty string -> Multiple None's
            {'key': [None, None]}
            >>> Server._flatten_query_args("notadict")  # Argument has to be a dict
            Traceback (most recent call last):
            ...
            TypeError: Argument 'args' is expected to be a (<class 'dict'>,), but is <class 'str'>

        """
        def _make_flat(item):
            if not isinstance(item, list):
                return item
            # item -> list
            if len(item) == 0:
                return None
            if len(item) == 1:
                return item[0] if item[0] else None  # Empty string -> None
            return [x if x else None for x in item]

        Validator.is_instance(dict, args=args)
        res = dict()
        for key, val in args.items():
            res[key] = _make_flat(val)
        return res



import os
import psutil

from . import Polling


class Stats (Polling):
    """
    Periodically emits stats about the host system, like cpu_use, memory_use, swap_use, ...
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def _root_path():
        return os.path.abspath(os.sep)

    @classmethod
    def _disk_usage(cls):
        try:
            return psutil.disk_usage(cls._root_path()).percent
        except FileNotFoundError:  # pragma: no cover
            return 0.0

    @staticmethod
    def _cpu_temp():
        try:
            # Raspberry
            with open('/sys/class/thermal/thermal_zone0/temp') as tz:
                return round(float(tz.read()) * 0.001, 1)
        except FileNotFoundError:  # pragma: no cover
            return 0.0

    def poll(self):
        l1m, l5m, l15m = os.getloadavg()
        return {
            'cpu_count': psutil.cpu_count(),
            'cpu_freq': psutil.cpu_freq().current,
            'cpu_use': psutil.cpu_percent(),
            'cpu_temp': self._cpu_temp(),
            'memory_use': psutil.virtual_memory().percent,
            'swap_use': psutil.swap_memory().percent,
            'disk_use': self._disk_usage(),
            'load_1m': l1m,
            'load_5m': l5m,
            'load_15m': l15m
        }



import multiprocessing as proc
import time
from abc import abstractmethod

from schedule import Scheduler

from .. import Plugin
from ...utils import auto_str_ignore, parse_duration_literal, StopCycleError, interruptible_sleep, try_parse_bool


@auto_str_ignore(['stopped', '_stopped', 'on_payload'])
class PullBase(Plugin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._stopped = proc.Event()

    @property
    def stopped(self):
        return self._stopped.is_set()

    @abstractmethod
    def pull(self):
        raise NotImplementedError()  # pragma: no cover

    def on_payload(self, payload):
        pass  # pragma: no cover

    def notify(self, payload):
        self.on_payload(self, payload)

    def stop(self):
        self._stopped.set()

    def _sleep(self, sleep_time=10):
        def callback():
            if self.stopped:
                raise StopCycleError()
        interruptible_sleep(sleep_time, callback, interval=0.5)


class PollingError(Exception):
    pass


class StopPollingError(Exception):
    """Raise this exception in child classes if you want to stop the polling scheduler."""
    pass


@auto_str_ignore(['_scheduler'])
class Polling(PullBase):
    __prefix__ = 'poll'

    def __init__(self, interval=60, instant_run=False, **kwargs):
        super().__init__(**kwargs)
        # polling interval in seconds
        self._interval = parse_duration_literal(interval)
        self._scheduler = None
        self._instant_run = try_parse_bool(instant_run, False)

    def pull(self):
        self._scheduler = Scheduler()
        self._scheduler.every(self._interval).seconds.do(self.run_schedule)

        if self._instant_run:
            self._scheduler.run_all()

        while not self.stopped:
            self._scheduler.run_pending()
            time.sleep(0.5)

    def run_schedule(self):
        try:
            payload = self.poll()
            self.notify(payload)
        except StopPollingError:
            self.stop()
        except:  # pragma: no cover
            import traceback
            self.logger.error("[{name}]\n{error}".format(name=self.name, error=traceback.format_exc()))

    @abstractmethod
    def poll(self):
        raise NotImplementedError()  # pragma: no cover



import random

from collections import defaultdict


class DHTMock:
    DHT22 = "dht22"
    DHT11 = "dht11"
    AM2302 = "am2302"

    @staticmethod
    def read_retry(sensor, pin):
        return round(random.uniform(1, 100), 2), round(random.uniform(8, 36), 2)  # pragma: no cover


class GPIOMock:  # pragma: no cover
    BCM = "bcm"
    IN = "in"
    RISING = "rising"
    FALLING = "falling"
    BOTH = "both"

    MODES = dict()
    CALLBACKS = defaultdict(list)
    SETUPS = set()

    @classmethod
    def setmode(cls, mode):
        pass

    @classmethod
    def setup(cls, channel, in_out):
        if channel in cls.SETUPS:
            raise RuntimeError("You called setup more than one time for channel '{}'".format(channel))
        cls.SETUPS.add(channel)

    @classmethod
    def cleanup(cls):
        pass

    @classmethod
    def add_event_detect(cls, channel, mode, callback=None, bouncetime=None):
        if channel not in cls.SETUPS:
            raise RuntimeError("You did not setup the channel '{}' before activating event detection".format(channel))
        cls.MODES[channel] = mode
        if not callback:
            return
        if channel in cls.CALLBACKS:
            raise RuntimeError("You already added a callback to this channel")
        cls.CALLBACKS[channel].append((callback, mode))

    @classmethod
    def add_event_callback(cls, channel, callback):
        if channel not in cls.MODES:
            raise RuntimeError("You have to enable the event detection first")
        mode = cls.MODES[channel]
        cls.CALLBACKS[channel].append((callback, mode))

    @classmethod
    def remove_event_detect(cls, channel):
        cls.CALLBACKS.pop(channel, None)
        cls.MODES.pop(channel, None)

    @classmethod
    def clear(cls):
        cls.MODES.clear()
        cls.CALLBACKS.clear()
        cls.SETUPS.clear()

    @classmethod
    def fire_event(cls, channel, mode):
        for cb, cb_mode in cls.CALLBACKS[channel]:
            if mode == cb_mode or cb_mode == cls.BOTH:
                cb(channel)



import os

from .config import load_config
from .engines import Engine, AdvancedRetryHandler
from .engines.thread import ThreadEngine
from .models import TaskSet, TaskModel, UDFModel, tasks_to_str
from .selector import PayloadSelector
from .utils import Loggable
from .validator import Validator


class Application(Loggable):
    def __init__(self, tasks: TaskSet):
        self._tasks = tasks
        self._engine = None

    def start(self):
        self._engine.run(self._tasks)

    def bind(self, engine: Engine):
        Validator.is_instance(Engine, engine=engine)
        self._engine = engine

    @classmethod
    def from_file(cls, file_path: str):
        udfs, engine, task_cfg = load_config(file_path)
        base_path = os.path.dirname(file_path)
        tasks = {
            task.name: TaskModel.from_dict(task, base_path) for task in task_cfg
        }
        if engine is None:
            # Backward compatibility
            engine = ThreadEngine(queue_worker=3, retry_handler=AdvancedRetryHandler())
        if udfs is not None:
            udfs = [UDFModel.from_config(udf) for udf in udfs]
            PayloadSelector.instance.register_udfs(udfs)

        from pprint import pformat
        cls.logger.info("UDFs\n{}".format(pformat(udfs)))
        cls.logger.info("Engine\n{}".format(engine))
        cls.logger.info("Configured tasks\n{}".format(tasks_to_str(tasks)))
        app = Application(tasks=tasks)
        app.bind(engine=engine)
        return app



import sys
import time

from . import PullBase, Polling
from ...validator import Validator


class Count(PullBase):
    """
    Emits every `wait` seconds a counting value which runs from `from_cnt` to `to_cnt`.
    If `to_cnt` is None will to count to infinity.

    Args:
        from_cnt (int): Where to start counting.
        to_cnt (int or None): Where to stop counting (if None will count to infinity).
        wait (float): Will wait that amount between counting emits.

    Returns:
        The `on_payload` callback will pass the current count as payload.

    Example configuration:

        name: count
        pull:
          plugin: pnp.plugins.pull.simple.Count
          args:
            from_cnt: 0
            to_cnt: 10
            wait: 5
        push:
          plugin: pnp.plugins.push.simple.Echo


    """

    def __init__(self, from_cnt=0, to_cnt=None, wait=5, **kwargs):
        super().__init__(**kwargs)
        self.from_cnt = int(from_cnt)
        self.to_cnt = to_cnt and int(to_cnt)
        self.wait = float(wait)

    def pull(self):
        for i in range(self.from_cnt, self.to_cnt or sys.maxsize):
            self._sleep(self.wait)
            self.notify(i)
            if self.stopped:
                break


class Repeat(PullBase):
    """
    Emits every `wait` seconds the same `repeat`.

    Args:
        repeat (any): Will repeat this every emit.
        wait (float): Will wait that amount between counting emits.

    Returns:
        The `on_payload` callback will pass the repeated object as payload.

    Example configuration:

    name: repeat
    pull:
      plugin: pnp.plugins.pull.simple.Repeat
      args:
        repeat: "Hello World"  # Repeats 'Hello World'
        wait: 1  # Every second
    push:
      plugin: pnp.plugins.push.simple.Echo

    """

    def __init__(self, repeat, wait=5, **kwargs):
        super().__init__(**kwargs)
        self.repeat = repeat
        self.wait = float(wait)

    def pull(self):
        while not self.stopped:
            self._sleep(self.wait)
            self.notify(self.repeat)


class Infinite(PullBase):
    """Just for demonstration purposes. DO NOT USE!"""

    def __init__(self, **kwargs):  # pragma: no cover
        super().__init__(**kwargs)

    def pull(self):  # pragma: no cover
        while True:
            try:
                time.sleep(0.5)
            except:
                pass


class CustomPolling(Polling):
    """
    Calls the specified callable every `interval`. The result of the callable is simply returned.

    Args:
        scheduled_callable (callable): Custom function to execute every `interval`.

    Returns:
        The `on_payload` callback will pass anything that the scheduled_callable has returned.

    Example configuration:

        Does not work with yaml or json configurations so far.
    """
    def __init__(self, scheduled_callable, **kwargs):
        super().__init__(**kwargs)
        self.scheduled_callable = scheduled_callable
        Validator.is_function(scheduled_callable=self.scheduled_callable)

    def poll(self):
        return self.scheduled_callable()



import time

from . import PullBase
from .. import load_optional_module
from ...utils import make_list, load_file, FILE_MODES, Debounce
from ...validator import Validator


class FileSystemWatcher(PullBase):
    """
    Watches the given directory for changes like created, moved, modified and deleted files. If `ignore_directories` is
    set to False, then directories will be reported as well.

    Per default will recursively report any file that is touched, changed or deleted in the given path. The
    directory itself or subdirectories will be object to reporting too, if `ignore_directories` is set to False.

    Args:
        path (str): The path to track for events.
        recursive (bool): If set to True, any subfolders of the given path will be tracked too.
        patterns (str or list): Any file pattern (e.g. *.txt or [*.txt, *.md].
        ignore_patterns (str or list): Any patterns to ignore (specify like argument patterns).
        ignore_directories (str): If set to True will send events for directories when file change.
        case_sensitive (bool): If set to True, any pattern is case_sensitive, otherwise it is case insensitive.
        events (str or list): The events to track. One or multiple of 'moved', 'deleted', 'created' and/or 'modified'.
        load_file (bool): If set to True the file will be loaded to the result.
        mode (str): Open mode of the file (only necessary when load_file is True). Can be text, binary or auto
            (guessing).
        base64 (bool): If set to True the loaded file will be converted to base64 (only applicable when
            load_file is True). `mode` will be automatically set to 'binary'.
        defer_modified (float): If set greater than 0, it will defer the sending of modified events for that amount.
            There might be multiple flushes of a file before it is written completely to disk.
            Without defer_modified each flush will raise a modified event.

    Returns:
        The callback `on_payload` will offer a payload that is a dictionary that represents the reported file
        system event. Example ->
            {
                'operation': 'modified',
                'source': '/tmp/abc.txt',
                'is_directory': False,
                'destination': None,  # Only non-None when operation = 'moved'
                'file': {  # Only present when load_file is True
                    'file_name': 'abc.txt',
                    'content': 'foo and bar',
                    'read_mode': 'text',
                    'base64': False
                }
            }

    Example configuration:

        name: watcher
        pull:
          plugin: pnp.plugins.pull.fs.FileSystemWatcher
          args:
            path: '/tmp'
            recursive: True
            patterns: '*.txt'
            ignore_directories: True
            case_sensitive: False
            events: [created, deleted]
            load_file: False
            mode: text
            base64: False
            defer_modified: 0
        push:
            plugin: pnp.plugins.push.simple.Echo
    """

    EXTRA = 'fswatcher'
    EVENT_TYPE_MOVED = 'moved'
    EVENT_TYPE_DELETED = 'deleted'
    EVENT_TYPE_CREATED = 'created'
    EVENT_TYPE_MODIFIED = 'modified'
    EVENT_TYPES = [EVENT_TYPE_MOVED, EVENT_TYPE_DELETED, EVENT_TYPE_CREATED, EVENT_TYPE_MODIFIED]

    def __init__(self, path, recursive=True, patterns=None, ignore_patterns=None, ignore_directories=False,
                 case_sensitive=False, events=None, load_file=False, mode='auto', base64=False, defer_modified=0.5,
                 **kwargs):
        super().__init__(**kwargs)
        self.path = path
        Validator.is_directory(path=self.path)
        self.recursive = bool(recursive)
        self.patterns = make_list(patterns)
        self.ignore_patterns = make_list(ignore_patterns)
        self.ignore_directories = bool(ignore_directories)
        self.case_sensitive = bool(case_sensitive)
        self.load_file = bool(load_file)
        self.base64 = bool(base64)
        self.events = make_list(events)
        Validator.subset_of(self.EVENT_TYPES, allow_none=True, events=self.events)
        self.mode = mode
        Validator.one_of(FILE_MODES, mode=self.mode)
        self.defer_modified = float(defer_modified)
        Validator.is_non_negative(allow_none=True, defer_modified=self.defer_modified)

    def pull(self):
        Observer = load_optional_module('watchdog.observers', self.EXTRA).Observer
        wev = load_optional_module('watchdog.events', self.EXTRA)

        that = self

        class EventHandler(wev.PatternMatchingEventHandler):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)
                self.dispatcher = {}

            @staticmethod
            def _read_file_from_event(event):
                if event.event_type != that.EVENT_TYPE_DELETED and not getattr(event, 'is_directory', False):
                    file_name = getattr(event, 'dest_path', None) or getattr(event, 'src_path', None)
                    if file_name is not None:
                        return load_file(file_name, that.mode, that.base64)
                return None

            def stop_dispatcher(self):
                # We cannot alter the dictionary during iteration - so we have to get all relevant debounces ...
                candidates = [debounce for _, debounce in self.dispatcher.items()]
                # ... and then loop it over outside the iterator context
                for debounce in candidates:
                    debounce.execute_now()

            def _deferred_notify(self, modified_file, payload):
                # Let's remove the dispatcher instance - it is done
                self.dispatcher.pop(modified_file, None)
                # trigger the actual notification event
                that.notify(payload)

            def _notify(self, event, payload):
                if event.event_type == that.EVENT_TYPE_MODIFIED:
                    # There might be multiple flushes of a file before it is completely written to disk
                    # Each flush will raise a modified event... Let's wait for more modified events...
                    if that.defer_modified <= 0:
                        that.notify(payload)
                    else:
                        modified_file = payload['source']
                        defer_fun = self.dispatcher.get(modified_file, None)
                        if defer_fun is None:
                            defer_fun = Debounce(self._deferred_notify, that.defer_modified)
                            self.dispatcher[modified_file] = defer_fun
                        defer_fun(modified_file, payload)
                        that.logger.debug("[{that.name}] Event of modified_file '{modified_file}' "
                                          "is deferred for {that.defer_modified}".format(**locals()))
                else:
                    defer_fun = self.dispatcher.get(payload['source'], None)
                    # There might be some deferred modified event... Lets check and send it to keep the correct
                    # sequence. Assuming the file is closed and finished by now...
                    if defer_fun is not None:
                        defer_fun.execute_now()
                    that.notify(payload)

            def on_any_event(self, event):
                if that.events is None or event.event_type in that.events:
                    payload = {
                        'operation': event.event_type,
                        'is_directory': getattr(event, 'is_directory', False),
                        'source': getattr(event, 'src_path', None),
                        'destination': getattr(event, 'dest_path', None)
                    }
                    that.logger.info("[{that.name}] Got {payload}".format(**locals()))
                    if that.load_file:
                        file_envelope = self._read_file_from_event(event)
                        if file_envelope is not None:
                            payload['file'] = file_envelope

                    self._notify(event, payload)

        observer = Observer()
        handler = EventHandler(
            patterns=self.patterns,
            ignore_patterns=self.ignore_patterns,
            ignore_directories=self.ignore_directories,
            case_sensitive=self.case_sensitive
        )
        observer.schedule(handler, self.path, recursive=self.recursive)
        observer.start()

        while not self.stopped:
            time.sleep(1)

        observer.stop()
        handler.stop_dispatcher()
        observer.join()



import copy
from abc import abstractmethod

from .. import Plugin
from ... import utils
from ...validator import Validator


class PushExecutionError(Exception):
    pass


class PushBase(Plugin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @abstractmethod
    def push(self, payload):
        raise NotImplementedError()  # pragma: no cover

    def _parse_envelope_value(self, name, envelope=None, parse_fun=None, instance_lookup_fun=None):
        """
        Parse the envelope for the given `name`. If present in the envelope the extracted value will be tested / parsed
        by the given `parse_fun`. If no `parse_fun` is explicitly given, it will be determined if this instance
        provides a `parse_<name>`, `_parse_<name>` or a `__parse_<name>`. If yes the function is called; otherwise the
        value is returned as is.

        If the value is not present in the envelope the current instance will be probed for `<name>`, `_<name>` or
        `__<name>`. If an instance variable is present it will be returned unvalidated / unparsed (we assume that
        happened previously). If no variable is present simply `None` will be returned.
        Args:
            name: Name of the attribute to lookup in the envelope / instance.
            envelope: Envelope of the payload.
            parse_fun: Custom function to validate / parse. If not given it will be determined automagically.
            instance_lookup_fun: Custom function to perform instance variable lookups. If not given a reasonably default
                will be used.
        """
        Validator.is_instance(str, name=name)
        Validator.is_instance(dict, allow_none=True, envelope=envelope)
        Validator.is_function(allow_none=True, parse_fun=parse_fun)
        Validator.is_function(allow_none=True, instance_lookup_fun=instance_lookup_fun)

        lookups = utils.make_public_protected_private_attr_lookup(name, as_dict=True)

        if envelope is None or name not in envelope:
            return (instance_lookup_fun(name)
                    if instance_lookup_fun is not None
                    else utils.instance_lookup(self, list(lookups.values())))

        val = envelope[name]
        try:
            if parse_fun is None:
                public_attr_name = lookups['public']
                parse_fun_names = utils.make_public_protected_private_attr_lookup('parse_' + public_attr_name)
                parse_fun = utils.instance_lookup(self, parse_fun_names)
                if parse_fun is None:
                    return val
            return parse_fun(val)
        except (ValueError, TypeError):
            import traceback
            self.logger.error("[{name}] Cannot parse value for '{attr}' from envelope. Error is\n{error}".format(
                name=self.name, attr=name, error=traceback.format_exc()))
            return (instance_lookup_fun(name)
                    if instance_lookup_fun is not None
                    else utils.instance_lookup(self, list(lookups.values())))

    @staticmethod
    def envelope_payload(payload):
        """
        A payload might be enveloped (but actually it even may not). The actual payload will be available via
        data or payload inside the dictionary. All other keys are the envelope.
        Args:
            payload: Payload (which might be enveloped or not).

        Returns:
            A tuple consisting of envelope (might be None) and the real payload.

        Examples:

            >>> PushBase.envelope_payload(1234)  # No envelope -> payload as is
            ({}, 1234)
            >>> PushBase.envelope_payload({'key1': 'val1', 'key2': 'val2'}) == ({}, {'key1': 'val1', 'key2': 'val2'})
            True
            >>> PushBase.envelope_payload({'data': 1234})  # Envelope with the data key only
            ({}, 1234)
            >>> PushBase.envelope_payload({'payload': 1234})  # Works with 'payload' as well
            ({}, 1234)
            >>> PushBase.envelope_payload({'envelope1': 'envelope', 'payload': 1234})  # With actual envelope
            ({'envelope1': 'envelope'}, 1234)
        """
        payload = copy.deepcopy(payload)

        if not isinstance(payload, dict):
            return {}, payload

        if 'data' in payload:
            real_payload = payload.pop('data')
        elif 'payload' in payload:
            real_payload = payload.pop('payload')
        else:
            return {}, payload  # There is no envelope

        return payload, real_payload  # We popped the real payload -> all what is left in payload is the envelope



import re

import requests
import schema
from requests.auth import HTTPBasicAuth

from . import Polling, PollingError
from .http import Server
from ...utils import auto_str_ignore, try_parse_int_float_str
from ...validator import Validator


@auto_str_ignore(['password'])
class ZwayPoll(Polling):
    """
    Pulls the specified json content from the zway rest api. The content is specified by the url, e.g.
    `http://<host>:8083/ZWaveAPI/Run/devices` will pull all devices and serve the result as a json.

    Specify the polling interval by setting the argument `interval`. User / password combination is required when
    your api is protected against guest access (by default it is).

    Use multiple pushes and the related selectors to extract the required content like temperature readings (see
    the example configuration for guidance).

    All arguments (`url`, `user` and `password`) can be automatically injected via environment variables.
    * ZWAY_URL
    * ZWAY_USER
    * ZWAY_PASSWORD

    Args:
        url (str): The zway api url to poll.
        user (str): The user for authentication to the api.
        password (str): The related password.

    Returns:
        The `on_payload` callback will pass the result of specified url.

    Example configuration:

        name: zway
        pull:
          plugin: pnp.plugins.pull.ZwayPoll
          args:
            url: http://<host>:8083/ZWaveAPI/Run/devices  # Retrieve all devices
            user: admin
            password: secret
            interval: 5m  # Poll the api every 5 minutes
        push:
          - plugin: pnp.plugins.push.Echo
            selector: payload[19].instances[0].commandClasses[49].data[1].val.value  # temp. of fibaro motion sensor
          - plugin: pnp.plugins.push.Echo
            selector: payload[9].instances[0].commandClasses[128].data.last.value  # Setpoint of heater

    """

    __prefix__ = 'zway'

    def __init__(self, url, user, password, **kwargs):
        super().__init__(**kwargs)
        self.url = str(url)
        self.user = str(user)
        self.password = str(password)

    def poll(self):
        self.logger.debug("[{name}] Polling url '{url}'".format(name=self.name, url=self.url))
        reply = requests.get(self.url, auth=HTTPBasicAuth(self.user, self.password))
        if reply.status_code != requests.codes.ok:
            raise PollingError("Code {code}: {error}".format(code=reply.status_code, error=reply.text))

        try:
            return reply.json()
        except ValueError:
            # No valid json, try text
            return reply.text


class ZwayReceiver(Server):
    """
    Setups a http server to process incoming GET-requests from the
    Zway-App [`HttpGet`](https://github.com/hplato/Zway-HTTPGet/blob/master/index.js).

    Besides the arguments noted below the component will accept any arguments that `pnp.plugins.pull.http.Server`
    would accept.

    Args:
        url_format (str): The url_format that is configured in your HttpGet App. If you configured
            `http://<ip>:<port>/set?device=%DEVICE%&state=%VALUE%` (default of the App), you basically have to copy
            the path component `set?device=%DEVICE%&state=%VALUE%` to be your `url_format`.
        device_mapping (Union[Dict[Str, Str], Dict[Str, Dict]], optional): A mapping to map the somewhat cryptic
            virtual device names to human readable ones. Default is None, which means that no mapping will be performed.
            Two ways possible:
                1. Ordinary mapping from virtual device name -> alias.
                2. Enhanced mapping from virtual device name to dictionary with additional properties. One property
                has to be alias.
        ignore_unknown_devices (bool, optional): If set to True all incoming requests that are associated with an
            device that is not part of the mapping will be ignored. Default is False.<br/>

    Returns:

        Given the url_format `%DEVICE%?value=%VALUE%`, the url `http://<ip>:<port>/set?vdevice1?value=5.5` and
        the device_mapping `vdevice1 -> alias of vdevice1` the emitted message will look like this:

        ```yaml
        {
            'device_name': 'alias of vdevice1',
            'raw_device': 'vdevice1'
            'value': 5.5,
            'props': {}
        }
        ```
    """
    MAPPING_SCHEMA = schema.Schema({
        schema.Optional(str): schema.Or(
            {
                'alias': schema.Use(str),
                schema.Optional(str): object
            },
            schema.Use(lambda val: dict(alias=val, props={}))
        )
    })

    DEVICE_REGEX = r'[a-zA-Z0-9_\-]+'
    VALUE_REGEX = r'[a-zA-Z0-9\.,]+'
    MODE_AUTO = 'auto'
    MODE_MAPPING = 'mapping'
    MODE_BOTH = 'both'
    VALID_MODES = [MODE_AUTO, MODE_MAPPING, MODE_BOTH]
    VDEV_REGEX = r'ZWayVDev_zway_(?P<device_id>[0-9]+)\-(?P<instance>[0-9]+)\-(?P<cclass>[0-9]+)(\-(?P<mode>[0-9]+))?'

    AUTO_MAPPING = {
        '37': 'switch',  # Switch binary
        '38': 'level',  # Switch multilevel
        '48': {'1': 'motion'},  # Sensor binary
        '49': {'1': 'temperature', '3': 'illumination', '4': 'power'},  # Sensor multilevel
        '50': {'0': 'consumption'},  # Meter
        '67': {'1': 'setpoint'},  # Thermostat
        '128': 'battery'
    }

    @staticmethod
    def _make_regex(url_format):
        escaped_url = re.escape(url_format)
        device_group = "(?P<device>{})".format(ZwayReceiver.DEVICE_REGEX)
        value_group = "(?P<value>{})".format(ZwayReceiver.VALUE_REGEX)
        return re.compile('.*' + escaped_url.replace(r'\%DEVICE\%', device_group).replace(r'\%VALUE\%', value_group)
                          .replace(r'%DEVICE%', device_group).replace(r'%VALUE%', value_group))

    @staticmethod
    def _safe_get_group(match, group_name):
        try:
            return match.group[group_name]
        except IndexError:
            return None

    @classmethod
    def _auto_map(cls, cclass, mode):
        descr = cls.AUTO_MAPPING.get(str(cclass))
        if not descr:
            return None
        if isinstance(descr, dict) and mode:
            descr = descr.get(str(mode))
        return descr

    def __init__(self, url_format="/set?device=%DEVICE%&state=%VALUE%", device_mapping=None,
                 ignore_unknown_devices=False, mode='mapping', vdev_regex=None, **kwargs):
        # Zway module HttpGet only supports GET calls
        kwargs['allowed_methods'] = 'GET'
        super().__init__(**kwargs)
        self.ignore_unknown_devices = bool(ignore_unknown_devices)

        Validator.is_instance(str, url_format=url_format)
        if not url_format.startswith('/'):
            url_format = '/' + url_format
        if '?' not in url_format:
            url_format = url_format + '?'
        self.url_format = url_format
        self.url_format_regex = self._make_regex(url_format)

        if device_mapping is None:
            device_mapping = {}
        self.device_mapping = self.MAPPING_SCHEMA.validate(device_mapping)

        Validator.one_of(self.VALID_MODES, False, mode=mode)
        self.mode = mode

        if mode in [self.MODE_BOTH, self.MODE_AUTO]:
            self._vdev_regex = vdev_regex or self.VDEV_REGEX
            self._vdev_regex_compiled = re.compile(self.VDEV_REGEX)

    def notify(self, payload):
        full_path = payload['full_path']
        match = self.url_format_regex.match(full_path)
        if match is None:
            self.logger.warning("Cannot extract device_name and/or value from '{}' with regex '{}'".format(
                full_path, self.url_format_regex
            ))
            return
        raw_device = match.group('device')
        value = match.group('value')

        device_map = None
        if self.mode in [self.MODE_MAPPING, self.MODE_BOTH]:
            device_map = self.device_mapping.get(raw_device)

        if not device_map and self.mode in [self.MODE_AUTO, self.MODE_BOTH]:
            match = self._vdev_regex_compiled.match(raw_device)

            if match is not None:
                device_id, cclass, mode = match.group('device_id'), match.group('cclass'), match.group('mode')
                type_of_device = self._auto_map(cclass, mode)
                if not type_of_device and self.ignore_unknown_devices:
                    self.logger.debug("Got device '{raw_device}' but it is unknown to the automapper".format(
                        **locals()))
                    return
                device_map = {
                    'alias': device_id,
                    'command_class': cclass,
                    'mode': mode,
                    'type': type_of_device or (cclass + '-{}'.format(mode) if mode else '')
                }
            else:
                self.logger.warning("Cannot extract device_name and/or value from '{}' with regex '{}'".format(
                    raw_device, self._vdev_regex
                ))

        if device_map is None:
            if self.ignore_unknown_devices:
                self.logger.debug("Got device '{raw_device}' but it is ignored".format(**locals()))
                return  # Device is not mapped and should be ignored-> abort
            device_map = dict(alias=raw_device)  # If it isn't ignored just take the raw name

        device_map = device_map.copy()
        alias = device_map.pop('alias')
        super().notify(dict(device_name=alias, props=device_map, raw_device=raw_device,
                            value=try_parse_int_float_str(value)))



import multiprocessing as proc
import os

from .parallel import StoppableRunner, StoppableWorker, ParallelEngine


class ProcessStoppableRunner(StoppableRunner):
    """Runner implementation using multiprocessing thread-like interface."""
    @staticmethod
    def _make_shutdown_event():
        return proc.Event()

    @staticmethod
    def get_ident():
        return os.getppid()

    def start(self):
        if self._runner:
            raise RuntimeError("Runner is already started and cannot restart...")
        self._runner = proc.Process(target=self._start_pull)
        self._runner.start()

    def terminate(self):
        self._assert_runner()
        self._runner.terminate()


class ProcessWorker(StoppableWorker):
    """Worker implementation using multiprocessing thread-like interface."""
    @staticmethod
    def get_ident():
        return os.getppid()

    def start(self):
        if self._worker:
            raise RuntimeError("Worker is already started and cannot restart...")
        self._worker = proc.Process(target=self._process_queue)
        self._worker.start()

    def terminate(self):
        self._worker.terminate()


class ProcessEngine(ParallelEngine):
    """Engine implementation using multiprocessing thread-like interface."""
    def _make_worker(self):
        return ProcessWorker(self.queue, self.stop_working_item)

    def _make_runner(self, task, retry_handler):
        return ProcessStoppableRunner(task, self.queue, retry_handler=retry_handler)

    def _make_queue(self):
        return proc.Queue()

    def _make_shutdown_event(self):
        return proc.Event()

    def _stop_worker(self):
        # We do not use the stop_working_item here... We ungentle terminate the worker
        for worker in self.worker:
            self.logger.debug("Terminating worker: {}".format(worker))
            worker.terminate()



from abc import abstractmethod

from .. import Plugin


class UserDefinedFunction(Plugin):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def __call__(self, *args, **kwargs):
        return self.action(*args, **kwargs)

    @abstractmethod
    def action(self, *args, **kwargs):
        raise NotImplementedError()



import logging
import re
from abc import abstractmethod
from collections import defaultdict, Counter
from functools import partial

from . import PullBase
from ...utils import (make_list, try_parse_int, auto_str, auto_str_ignore, Debounce, parse_duration_literal,
                      Singleton, Loggable)
from ...validator import Validator

logger = logging.getLogger(__name__)


CONST_AVAILABLE_GPIO_PINS = list(range(2, 28))
CONST_RISING_OPTIONS = ["rise", "rising"]
CONST_FALLING_OPTIONS = ["fall", "falling"]
CONST_SWITCH_OPTIONS = ["button", "switch"]
CONST_MOTION_OPTIONS = ["motion"]

CONST_RISING = "rising"
CONST_FALLING = "falling"
CONST_SWITCH = "switch"
CONST_MOTION_ON = "motion_on"
CONST_MOTION_OFF = "motion_off"


class Watcher(PullBase):
    __prefix__ = 'gpio'

    EXTRA = 'gpio'

    def __init__(self, pins, default=CONST_RISING, **kwargs):
        super().__init__(**kwargs)
        self._mode_default = default
        Validator.one_of(CONST_RISING_OPTIONS + CONST_FALLING_OPTIONS + CONST_SWITCH_OPTIONS + CONST_MOTION_OPTIONS,
                         mode_default=self._mode_default)
        self._pins = [Callback.from_str(pin_str, default=default) for pin_str in make_list(pins)]
        _without_duplicate = set(self._pins)
        if len(_without_duplicate) != len(self._pins):
            diff = list((Counter(self._pins) - Counter(_without_duplicate)).elements())
            self.logger.warning("[{self.name}] You provided duplicate gpio pin configurations. Will ignore '{diff}'"
                                .format(**locals()))
            self._pins = _without_duplicate

    def _universal_callback(self, gpio_pin, event):
        self.logger.info("[{self.name}] GPIO '{gpio_pin}' raised event '{event}'".format(**locals()))
        self.notify(dict(gpio_pin=gpio_pin, event=event))

    def pull(self):
        GPIO = GPIOAdapter()

        try:
            for pin in self._pins:
                pin.run(self._universal_callback)
            GPIO.apply()
            while not self.stopped:
                self._sleep()
        finally:
            for pin in self._pins:
                pin.stop()
            GPIO.cleanup()


class GPIOAdapter (Singleton, Loggable):
    def __init__(self):
        self.logger.debug("GPIOWrapper __init__ called")
        self.GPIO = self._load_gpio_package()
        self.GPIO.setmode(self.GPIO.BCM)
        self._mchannel = dict()
        self._callbacks = defaultdict(list)

    def add_event_detect(self, channel, edge, callback, bouncetime=None):
        self._callbacks[channel].append((edge, callback, bouncetime))

    def apply(self):
        def safe_max(l):
            if not len(l):
                return None
            return max(l)

        for channel, ev in self._callbacks.items():
            self.GPIO.setup(channel, self.GPIO.IN)
            # If multiple bounce times are specified, use max
            max_bouncetime = safe_max([bouncetime for _, _, bouncetime in ev if bouncetime])
            edges = list(set([edge for edge, _, _ in ev]))
            edge = edges[0] if len(edges) == 1 else self.GPIO.BOTH

            # We have to use both mode because the pin is configured to use rising and falling.
            # But the callback does not provide the edge, so we have to guess in those cases ;-)
            if edge == self.GPIO.BOTH:
                self.logger.debug("Channel '{channel}' does provide rising and falling - Using 'both'".format(
                    **locals()))
                self._mchannel[channel] = None

            gpio_edge = self._str_to_gpio_mode(edge)
            if max_bouncetime:
                self.logger.debug("Channel '{channel}' uses bouncetime of {max_bouncetime}ms. "
                                  "This will apply for all callbacks".format(**locals()))
                self.GPIO.add_event_detect(channel, gpio_edge, bouncetime=max_bouncetime)
            else:
                self.GPIO.add_event_detect(channel, gpio_edge)

            self.GPIO.add_event_callback(channel, self._dispatcher)

    def remove_event_detect(self, channel):
        self._mchannel.pop(channel, None)
        self._callbacks.pop(channel, None)
        return self.GPIO.remove_event_detect(channel)

    def cleanup(self):
        self._mchannel.clear()
        self._callbacks.clear()
        return self.GPIO.cleanup()

    def _dispatcher(self, channel):
        callbacks = self._callbacks.get(channel, None)
        if not callbacks:
            return
        is_multi_channel = channel in self._mchannel
        current_edge = None
        if is_multi_channel:
            last_state = self._mchannel[channel] or CONST_FALLING
            current_edge = CONST_RISING if last_state == CONST_FALLING else CONST_FALLING
            self._mchannel[channel] = current_edge

        for cb_edge, cb, _ in callbacks:
            if current_edge is None or cb_edge == current_edge:
                cb(channel)

    def _str_to_gpio_mode(self, s):
        if s == self.GPIO.BOTH:
            return s
        cs = str(s)
        if cs.lower() in CONST_RISING_OPTIONS:
            return self.GPIO.RISING
        if cs.lower() in CONST_FALLING_OPTIONS:
            return self.GPIO.FALLING
        raise ValueError("The given string '{}' is not a valid GPIO mode".format(cs))

    @staticmethod
    def _load_gpio_package():
        try:
            try:
                import RPi.GPIO as GPIO
            except ImportError:
                from RPi import GPIO  # Only works for test cases when using the mocked package
        except ImportError:  # pragma: no cover
            logger.warning("RPi.GPIO package is not available - Using mock")
            from ...mocking import GPIOMock as GPIO
        return GPIO


@auto_str(__repr__=True)
@auto_str_ignore(['_GPIO'])
class Callback:
    def __init__(self, gpio_pin):
        self.gpio_pin = self._str_to_gpio_pin(gpio_pin)

    @property
    def GPIO(self):
        return GPIOAdapter()

    @classmethod
    def from_str(cls, candidate, default=CONST_RISING):
        gpio_pin_str, _, mode_str = str(candidate).partition(':')
        if not mode_str:
            mode_str = default
        if any(mode_str.lower().startswith(aswitch) for aswitch in CONST_RISING_OPTIONS):
            return RisingCallback.load(gpio_pin_str, mode_str)
        if any(mode_str.lower().startswith(aswitch) for aswitch in CONST_FALLING_OPTIONS):
            return FallingCallback.load(gpio_pin_str, mode_str)
        if any(mode_str.lower().startswith(aswitch) for aswitch in CONST_SWITCH_OPTIONS):
            return SwitchCallback.load(gpio_pin_str, mode_str)
        if any(mode_str.lower().startswith(aswitch) for aswitch in CONST_MOTION_OPTIONS):
            return MotionCallback.load(gpio_pin_str, mode_str)
        raise ValueError("The specified pin mode '{}' is not supported".format(mode_str))

    def _intercept(self, channel, event, callback):
        callback(gpio_pin=channel, event=event)

    @abstractmethod
    def run(self, callback):
        raise NotImplementedError()  # pragma: no cover

    def stop(self):
        self.GPIO.remove_event_detect(self.gpio_pin)

    def _eqstr(self):
        # If we have multiple switches or motion for the same pin only one will applied
        # We basically account for the pin and the class only without additional params.
        return "{}({})".format(type(self).__name__, str(self.gpio_pin))

    @staticmethod
    def _parse_mode_str(mode_str, regex, *groups):
        r = re.compile(regex)
        m = r.match(mode_str)
        if not m:
            raise ValueError("The given mode '{}' is not a valid definition for a callback".format(mode_str))
        return [m.group(g) for g in groups]

    @staticmethod
    def _str_to_gpio_pin(s):
        pin = try_parse_int(s)
        if not pin or pin not in CONST_AVAILABLE_GPIO_PINS:
            lmin = min(CONST_AVAILABLE_GPIO_PINS)
            lmax = max(CONST_AVAILABLE_GPIO_PINS)
            raise ValueError("The given string '{}' is not a valid GPIO pin. "
                             "String needs to be convertible to int and between {} and {}".format(s, lmin, lmax))
        return pin

    def __hash__(self):
        return hash(str(self._eqstr()))

    def __eq__(self, other):
        if not hasattr(other, '_eqstr'):  # pragma: no cover
            return False
        return self._eqstr() == other._eqstr()


class RisingCallback(Callback):
    EDGE = CONST_RISING

    @classmethod
    def load(cls, gpio_pin_str, mode_str):
        return cls(gpio_pin_str)

    def run(self, callback):
        self.GPIO.add_event_detect(
            self.gpio_pin,
            self.EDGE,
            callback=partial(self._intercept, event=self.EDGE, callback=callback)
        )


class FallingCallback(RisingCallback):
    EDGE = CONST_FALLING


class SwitchCallback(Callback):
    EDGE = CONST_RISING
    BOUNCE_REGEX = r"^(switch|button)(\((?P<delay>\d+)\))?$"
    BOUNCE_GROUP = "delay"
    BOUNCE_DEFAULT = 500

    def __init__(self, gpio_pin, delay):
        super().__init__(gpio_pin)
        self.delay = int(delay)

    @classmethod
    def load(cls, gpio_pin_str, mode_str):
        delay, = cls._parse_mode_str(mode_str, cls.BOUNCE_REGEX, cls.BOUNCE_GROUP)
        return cls(gpio_pin_str, delay or cls.BOUNCE_DEFAULT)

    def run(self, callback):
        self.GPIO.add_event_detect(
            self.gpio_pin,
            self.EDGE,
            bouncetime=self.delay,
            callback=partial(self._intercept, event=CONST_SWITCH, callback=callback)
        )


class MotionCallback(RisingCallback):
    DELAY_REGEX = r"^motion(\((?P<delay>\d+[sSmMhHdDwW]?)\))?$"
    DELAY_GROUP = "delay"
    DELAY_DEFAULT = "30s"

    def __init__(self, gpio_pin, delay):
        super().__init__(gpio_pin)
        self.delay = parse_duration_literal(delay)
        self._debouncer = None

    @classmethod
    def load(cls, gpio_pin_str, mode_str):
        delay, = cls._parse_mode_str(mode_str, cls.DELAY_REGEX, cls.DELAY_GROUP)
        return cls(gpio_pin_str, delay or cls.DELAY_DEFAULT)

    def _motion_off(self, channel, callback):
        callback(gpio_pin=channel, event=CONST_MOTION_OFF)
        self._debouncer = None

    def _intercept(self, channel, event, callback):
        if event == CONST_RISING:
            if self._debouncer is None:
                callback(gpio_pin=channel, event=CONST_MOTION_ON)
                self._debouncer = Debounce(self._motion_off, self.delay)
            self._debouncer(channel, callback)

    def stop(self):
        if self._debouncer:
            self._debouncer.execute_now()



import copy
import multiprocessing
import threading
import time
from queue import Queue

from . import RetryHandler, PushExecutor, Engine, SimpleRetryHandler
from ..models import TaskModel, TaskSet
from ..utils import Loggable, StopCycleError, interruptible_sleep, auto_str, auto_str_ignore
from ..validator import Validator


@auto_str(__repr__=True)
@auto_str_ignore(ignore_list=["queue", "stopped", "_runner"])
class StoppableRunner(Loggable):
    """
    Base implementation of a parallel runner that executes a `pull`. By default threading is used.

    For other implementations you have basically to override / adapt:
    * _make_shutdown_event() -> Create an event that can signal if this task should stop. The event should provide
                                `is_set()` and `set()` methods (like the `Event` from `threading` does).
    * get_ident() -> Some unique identifier
    * start() -> Starts the runner. Start is async and will not block the current execution.
    * join(timeout=...) -> Wait for the runner to finish. Will block!
    * is_alive() -> Checks if the runner is currently running
    * terminate() -> Terminate the runner. Will only be called when the runner does not stop after `stop()` was called.
    """

    def __init__(self, task, queue, retry_handler):
        """
        Initializer.
        Args:
            task: Task to execute.
            queue: Queue to use for communication between runner and worker. Should provide `put(item)` method.
            retry_handler: Retry handler.
        """
        Validator.is_instance(TaskModel, task=task)
        self.task = task
        self.queue = queue
        self.stopped = self._make_shutdown_event()
        Validator.is_instance(RetryHandler, retry_handler=retry_handler)
        self.retry_handler = retry_handler
        self._runner = None

    @staticmethod
    def _make_shutdown_event():
        return threading.Event()

    @staticmethod
    def get_ident():
        return threading.get_ident()

    def start(self):
        if self._runner:
            raise RuntimeError("Runner is already started and cannot restart...")
        self._runner = threading.Thread(target=self._start_pull)
        self._runner.start()

    def join(self, timeout=None):
        self._assert_runner()
        self._runner.join(timeout=timeout)

    def is_alive(self):
        self._assert_runner()
        return self._runner.is_alive()

    def terminate(self):
        self.logger.warn("Method terminate() on basic StoppableRunner is not supported. The thread will NOT terminate.")

    def _assert_runner(self):
        if not self._runner:
            raise RuntimeError("Runner is is not started")

    def _sleep(self, sleep_time):
        def callback():
            if self.stopped.is_set():
                raise StopCycleError()
        # We couldn't just easily call time.sleep(self, retry_wait), cause we cannot react on the stopping
        # signal. So have to manually sleep in 0.5 second steps and check for stopping signal
        interruptible_sleep(sleep_time, callback, interval=0.5)

    def _start_pull(self):
        def on_payload(plugin, payload):
            # A task / inbound might have multiple pushes / outbounds associated
            for push in self.task.pushes:
                self.logger.debug("[Task-{thread}] Queing item '{item}' for push '{push}'".format(
                    thread=self.get_ident(),
                    item=payload,
                    push=push
                ))
                # We do not evaluate the selector here. We let this handle the worker process. why?
                # Cause the selector might take some time... and we do not need a try... catch... here
                # Much can go wrong with the selector, if it does it will only cause issues with a single push
                self.queue.put((payload, push))

        self.task.pull.instance.on_payload = on_payload

        while not self.stopped.is_set():
            try:
                self.task.pull.instance.pull()
                if not self.stopped.is_set():
                    # Bad thing... Pulling exited unexpectedly
                    self.logger.error("[Task-{thread}] Pulling of '{pull}' exited unexpectedly".format(
                        thread=self.get_ident(),
                        pull=self.task.pull.instance,
                    ))
            except KeyboardInterrupt:
                self.logger.debug("[Task-{thread}] Pulling of '{pull}' hit a keyboard interrupt".format(
                    thread=self.get_ident(),
                    pull=self.task.pull.instance
                ))
            except:   # pylint: disable=broad-except
                # Bad thing... Pulling exited with exception
                import traceback
                self.logger.error("[Task-{thread}] Pulling of '{pull}' raised an error\n{error}".format(
                    thread=self.get_ident(),
                    pull=self.task.pull.instance,
                    error=traceback.format_exc()
                ))
            finally:
                if not self.stopped.is_set():
                    directive = self.retry_handler.handle_error()
                    if directive.abort:
                        self.logger.error("[Task-{thread}] Pulling of '{pull}' exited due to retry limitation".format(
                            thread=self.get_ident(),
                            pull=self.task.pull.instance,
                        ))
                        self.stopped.set()
                    else:
                        self._sleep(directive.wait_for)

    def stop(self):
        self.logger.info("[Task-{thread}] Got stopping signal: '{task}'".format(
            thread=self.get_ident(),
            task=self.task.name
        ))
        self._assert_runner()
        self.stopped.set()
        self.task.pull.instance.stop()


@auto_str(__repr__=True)
class StoppableWorker(Loggable):
    """
    Base implementation of a worker that processes fetches items from the queue and delegates it to a push for
    further execution. By default threading is used.

    To stop the worker you should put the `stop_working_item` on the queue.

    For other implementations you have basically to override / adapt:
    * get_ident() -> Some unique identifier
    * start() -> Starts the worker. Start is async and will not block the current execution.
    * join(timeout=...) -> Wait for the worker to finish. Will block!
    * is_alive() -> Checks if the worker is currently running
    * terminate() -> Terminate the worker. Will only be called when the worker does not stop after receiving the
                     `stop_working_item`.
    """
    def __init__(self, queue, stop_working_item):
        """

        Args:
            queue: Queue to fetch processable items from. Should provide `get()` method.
            stop_working_item: Stop the worker when this item hits the queue.
        """
        self.queue = queue
        self.stop_working_item = stop_working_item
        self._worker = None

    @staticmethod
    def get_ident():
        return threading.get_ident()

    def start(self):
        if self._worker:
            raise RuntimeError("Worker is already started and cannot restart...")
        self._worker = threading.Thread(target=self._process_queue)
        self._worker.start()

    def join(self, timeout=None):
        self._assert_runner()
        self._worker.join(timeout=timeout)

    def is_alive(self):
        self._assert_runner()
        return self._worker.is_alive()

    def terminate(self):
        self.logger.warn("Method terminate() on basic StoppableWorker is not supported. The thread will NOT terminate.")

    def _assert_runner(self):
        if not self._worker:
            raise RuntimeError("Runner is is not started")

    def _process_queue(self):
        while True:
            try:
                payload = self.queue.get()
                if payload is self.stop_working_item:
                    self.logger.info("[Worker-{thread}] Got stopping signal".format(
                        thread=threading.get_ident()))
                    # Notify other worker as well
                    self.queue.put(self.stop_working_item)
                    break
                try:
                    # We assume payload is actual payload, push
                    payload, push = payload
                    PushExecutor().execute(
                        id=threading.get_ident(),
                        payload=payload,
                        push=push,
                        result_callback=lambda res, dep: self.queue.put((res, dep))
                    )
                finally:
                    # self.queue.task_done()
                    pass
            except KeyboardInterrupt:
                pass
            except:  # pylint: disable=broad-except
                import traceback
                self.logger.error("\n{}".format(traceback.format_exc()))


@auto_str_ignore(ignore_list=["stop_working_item", "queue", "shutdown", "runner", "worker"])
class ParallelEngine(Engine):
    def __init__(self, queue_worker=3, retry_handler=None, stop_timeout=5):
        super().__init__()
        self.queue_worker = max(int(queue_worker), 1)
        self.retry_handler = retry_handler
        if self.retry_handler is None:
            self.retry_handler = SimpleRetryHandler()
        Validator.is_instance(RetryHandler, retry_handler=self.retry_handler)
        self.stop_working_item = object()
        self.queue = self._make_queue()
        self.runner = []
        self.worker = []
        self.shutdown = self._make_shutdown_event()
        self.stop_timeout = Validator.cast_or_none(int, stop_timeout) or 5

    def _make_worker(self):
        return StoppableWorker(self.queue, self.stop_working_item)

    def _make_runner(self, task, retry_handler):
        return StoppableRunner(task, self.queue, retry_handler=retry_handler)

    def _make_queue(self):
        return Queue()

    def _make_shutdown_event(self):
        # We might run the Engine in an isolated process. So the multiprocessing.Event() is better suited.
        return multiprocessing.Event()

    def _run_queue_worker(self):
        n_worker = self.queue_worker
        for i in range(n_worker):
            t = self._make_worker()
            t.start()
            self.logger.info("[Worker-{thread}] Started ({i}/{cnt})".format(
                thread=t.get_ident(), i=i + 1, cnt=n_worker))
            self.worker.append(t)

    def _run_tasks(self, tasks):
        for task_name, task in tasks.items():
            t = self._make_runner(task, copy.deepcopy(self.retry_handler))
            t.start()
            self.logger.info("[Task-{thread}] Started for task '{task}'".format(thread=t.get_ident(), task=task.name))
            self.runner.append(t)

    def _stop_runner(self):
        for runner in self.runner:
            self.logger.debug("Stopping runner: {}".format(runner))
            runner.stop()

    def _stop_worker(self):
        self.logger.debug("Pushing stop_working_item on queue")
        self.queue.put(self.stop_working_item)

    def _stop_all(self):
        self._stop_runner()
        self._stop_worker()

    def _wait_for_all(self):
        for t in self.runner + self.worker:
            t.join(timeout=self.stop_timeout)
            if t.is_alive():
                self.logger.warn("{} did not respond to stop command. Terminating...".format(t))
                t.terminate()

    def _run(self, tasks: TaskSet):
        for _, t in tasks.items():
            if not isinstance(t, TaskModel):
                raise TypeError("All items of argument 'tasks' are expected to be a 'Task' instance")

        self._run_queue_worker()
        self._run_tasks(tasks)
        try:
            while not self.shutdown.is_set():
                time.sleep(1)
        except KeyboardInterrupt:
            pass
        self.logger.debug("Exited control loop. Stopping all threads gracefully")
        self._stop_all()
        self._wait_for_all()

    def _stop(self):
        self.shutdown.set()



from . import PushBase
from ...utils import auto_str_ignore


@auto_str_ignore(['password'])
class InfluxPush(PushBase):
    """
    Pushes the given `payload` to an influx database using the line `protocol`.
    You have to specify `host`, `port`, `user`, `password` and the `database`.

    The `protocol` is basically a string that will be augmented at push-time with data from the payload.
    E.g. {payload.metric},room={payload.location} value={payload.value} assumes that payload contains metric, location
    and value.

    Args:
        host (str): The host where the influxdb is running.
        port (int): The port where the influxdb service is listening on.
        user (str): Username to use for authentication.
        password (str): Related password.
        database (str): The database to write to.
        protocol (str): Line protocol template (augmented with payload-data).
            See https://docs.influxdata.com/influxdb/v1.5/write_protocols/line_protocol_tutorial/.

    Returns:
        For the ability to chain multiple pushes together the payload is simply returned as is.
    """
    __prefix__ = 'influx'

    def __init__(self, host, port, user, password, database, protocol, **kwargs):
        super().__init__(**kwargs)
        self.host = str(host)
        self.port = int(port)
        self.user = str(user)
        self.password = str(password)
        self.database = str(database)
        self.protocol = str(protocol)

    def push(self, payload):
        envelope, real_payload = self.envelope_payload(payload)

        points = [self.protocol.format(payload=real_payload)]
        self.logger.debug("[{name}] Writing '{points}' to influxdb".format(name=self.name, points=str(points)))

        from influxdb import InfluxDBClient
        client = InfluxDBClient(self.host, self.port, self.user, self.password, self.database)
        client.write(points, {'db': self.database}, 204, 'line')

        return payload  # Payload as is. With envelope (if any)



import logging
from datetime import datetime

import requests

from . import Polling, PollingError
from ...utils import safe_get, auto_str_ignore
from ...validator import Validator

logger = logging.getLogger(__name__)


def load_dht_package():
    try:
        import Adafruit_DHT as DHT
    except ImportError:  # pragma: no cover
        logger.warning("Adafruit_DHT package is not available - Using mock")
        from ...mocking import DHTMock as DHT
    return DHT


class DHT(Polling):
    """
    Periodically polls a dht11 or dht22 (aka am2302) for temperature and humidity readings.
    Polling interval is controlled by `interval`.

    Args:
        device (str): The device to poll (one of dht22, dht11, am2302)
        data_gpio (int): The data gpio port where the device operates on.

    Returns:
        The callback `on_payload` passes a dictionary containing `temperature` and `humidity`.

    Example configuration:

    name: dht
        pull:
          plugin: pnp.plugins.pull.sensor.DHT
          args:
            device: dht22  # Connect to a dht22
            data_gpio: 17  # DHT is connected to gpio port 17
            interval: 5m  # Polls the readings every 5 minutes
        push:
          - plugin: pnp.plugins.push.simple.Echo
            selector: payload.temperature  # Temperature reading
          - plugin: pnp.plugins.push.simple.Echo
            selector: payload.humidity  # Humidity reading

    """
    __prefix__ = 'dht'

    EXTRA = 'dht'

    def __init__(self, device='dht22', data_gpio=17, humidity_offset=0.0, temp_offset=0.0, **kwargs):
        super().__init__(**kwargs)
        valid_devices = ['dht11', 'dht22', 'am2302']
        self.device = str(device).lower()
        Validator.one_of(valid_devices, device=self.device)
        self.data_gpio = int(data_gpio)
        self.humidity_offset = float(humidity_offset)
        self.temp_offset = float(temp_offset)

    def poll(self):
        # Adafruit package is optional - import at the last moment
        DHT = load_dht_package()

        device_map = {'dht11': DHT.DHT11, 'dht22': DHT.DHT22, 'am2302': DHT.AM2302}
        humidity, temperature = DHT.read_retry(device_map[self.device], self.data_gpio)

        if humidity is None or temperature is None:
            raise PollingError("Failed to get '{}' @ GPIO {} readings...".format(self.device, str(self.data_gpio)))

        return {
            'humidity': round(float(humidity) + self.humidity_offset, 2),
            'temperature': round(float(temperature) + self.temp_offset, 2)
        }


@auto_str_ignore(ignore_list=["api_key"])
class OpenWeather(Polling):
    """
    Periodically polls weather data from the OpenWeatherMap api.

    Args:
        api_key (str): The api_key you got from OpenWeatherMap after registration.
        lat (float): Latitude. If you pass `lat`, you have to pass `lon` as well.
        lon (float): Longitude. If you pass `lon`, you have to pass `lat` as well.
        city_name (str): The name of your city. To minimize ambiguity use lat/lon or your country as a suffix,
            e.g. London,GB. You have to pass whether `city_name` or `lat/lon`.
        units (str on of (metric, imperial, kelvin)): Specify units for temperature and speed.
            imperial = fahrenheit + miles/hour, metric = celsius + m/secs, kelvin = kelvin + m/secs. Default is metric.
        tz (str, optional): Time zone to use for current time and last updated time. Default is your local timezone.

    Results:
        A dictionary containing the results from OpenWeatherMap. See the docs for more information.
    """
    __prefix__ = 'openweather'

    def __init__(self, api_key, lat=None, lon=None, city_name=None, units="metric", tz=None, **kwargs):
        super().__init__(**kwargs)
        self.api_key = str(api_key)
        self.lat = Validator.cast_or_none(float, lat)
        self.lon = Validator.cast_or_none(float, lon)
        self.city_name = Validator.cast_or_none(str, city_name)
        if self.city_name is None and not self._validate_lat_lon():
            raise ValueError("You have to pass city_name or lat and lon.")

        Validator.one_of(["metric", "imperial", "kelvin"], units=units)
        self.units = units
        self.tz = Validator.cast_or_none(str, tz)

        from pytz import timezone
        from tzlocal import get_localzone
        self._tz = get_localzone() if self.tz is None else timezone(self.tz)

    def poll(self):
        url = self._create_request_url()
        resp = requests.get(url)
        if 200 != resp.status_code:
            raise PollingError("GET of '{url}' failed with status code = '{resp.status_code}'".format(**locals()))

        raw_data = resp.json()

        return dict(
            temperature=safe_get(raw_data, "main", "temp"),
            pressure=safe_get(raw_data, "main", "pressure"),
            humidity=safe_get(raw_data, "main", "humidity"),
            cloudiness=safe_get(raw_data, "clouds", "all"),
            wind=safe_get(raw_data, "wind"),
            poll_dts=datetime.now(self._tz).isoformat(),
            last_updated_dts=datetime.fromtimestamp(safe_get(raw_data, "dt"), tz=self._tz).isoformat(),
            raw=raw_data
        )

    def _validate_lat_lon(self):
        return self.lat is not None and self.lon is not None

    def _create_request_url(self):
        url = ("http://api.openweathermap.org/data/2.5/weather?"
               "units={self.units}&"
               "APPID={self.api_key}")
        if self.lat is None and self.lon is None:
            url += "&q={self.city_name}"
        else:
            url += "&lat={self.lat}&lon={self.lon}"
        return url.format(**locals())



from . import PushBase
from ...utils import is_iterable_but_no_str, parse_duration_literal
from ...validator import Validator


class Echo(PushBase):
    """
    This push simply logs the `payload` via the `logging` module.

    Examples:

        >>> dut = Echo(name="echo_push")
        >>> dut.push("I will be logged")
        'I will be logged'
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def push(self, payload):
        envelope, real_payload = self.envelope_payload(payload)
        self.logger.info("[{self.name}] Got '{real_payload}' with envelope '{envelope}'".format(**locals()))
        return payload  # Payload as is. With envelope (if any)


class Nop(PushBase):
    """
    Executes no operation at all. A call to push(...) just returns the payload.
    This push is useful when you only need the power of the selector for dependent pushes.

    Nop = No operation OR No push ;-)

    Examples:

        >>> dut = Nop(name="nop_push")
        >>> dut.push('I will be returned unaltered')
        'I will be returned unaltered'
    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.last_payload = None

    def push(self, payload):
        self.last_payload = payload
        return payload


class Execute(PushBase):
    """
    Executes a command with given arguments in a shell of the operating system.

    Will return the exit code of the command and optionally the output from stdout and stderr.
    """
    def __init__(self, command, args=None, cwd=None, capture=True, timeout="5s", **kwargs):
        super().__init__(**kwargs)
        self._command = str(command) if command else None
        self._args = self._parse_args(args)
        self._cwd = str(cwd) if cwd else None
        Validator.is_directory(allow_none=True, cwd=self._cwd)
        self._capture = bool(capture)
        self._timeout = timeout and parse_duration_literal(timeout)

    def _parse_args(self, val):
        if not val:
            return None
        # Check for list type
        if is_iterable_but_no_str(val):
            return " ".join(val)
        return str(val)

    def _execute(self, args):
        def _output(br):
            return [line.strip('\n\r') for line in br]

        def _command_str():
            if not args:
                return self._command
            return "{self._command} {args}".format(**locals())

        import subprocess
        cmd_str = _command_str()
        self.logger.info("[{self.name}] Running command in shell: {cmd_str}".format(**locals()))
        p = subprocess.Popen(
            args=cmd_str,
            shell=True,
            cwd=self._cwd,
            universal_newlines=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        try:
            if self._timeout:
                p.wait(timeout=self._timeout)

            res = dict(return_code=p.returncode)
            if self._capture:
                res['stdout'] = _output(p.stdout)
                res['stderr'] = _output(p.stderr)
            return res
        finally:
            p.stdout.close()
            p.stderr.close()

    def push(self, payload):
        envelope, real_payload = self.envelope_payload(payload)
        args = self._parse_envelope_value('args', envelope)  # Override args via envelope

        return self._execute(args)



from . import PushBase
from .. import load_optional_module
from ...utils import auto_str_ignore, get_bytes
from ...validator import Validator


@auto_str_ignore(['api_key'])
class Dropbox (PushBase):
    __prefix__ = 'dropbox'

    EXTRA = 'dropbox'

    def __init__(self, api_key, target_file_name=None, create_shared_link=True, **kwargs):
        super().__init__(**kwargs)
        self.api_key = str(api_key)
        self.target_file_name = str(target_file_name)
        self.create_shared_link = bool(create_shared_link)

    @staticmethod
    def _sanitze_target_file_name(val):
        Validator.is_not_none(val=val)
        val = str(val)
        if not val.startswith('/'):
            return '/' + val
        return val

    @staticmethod
    def _make_raw_file_url(shared_url):
        try:
            import urlparse
            from urllib import urlencode
        except:  # For Python 3
            import urllib.parse as urlparse
            from urllib.parse import urlencode

        url_parts = list(urlparse.urlparse(shared_url))
        # Replace all query params with ?raw=1
        url_parts[4] = urlencode({'raw': '1'})
        return urlparse.urlunparse(url_parts)

    def push(self, payload):
        envelope, real_payload = self.envelope_payload(payload)
        # Override target_file_name via envelope
        target_file_name = self._sanitze_target_file_name(self._parse_envelope_value('target_file_name', envelope))

        # Upload file or stream to dropbox
        dropbox = load_optional_module('dropbox', self.EXTRA)
        dbx = dropbox.Dropbox(self.api_key)
        fcontent = get_bytes(real_payload)  # Might be a file or a stream
        self.logger.info("[{self.name}] Uploading file to dropbox '{target_file_name}'".format(**locals()))
        metadata = dbx.files_upload(fcontent, target_file_name, mode=dropbox.files.WriteMode("overwrite"))

        res = dict(
            name=metadata.name,
            id=metadata.id,
            content_hash=metadata.content_hash,
            size=metadata.size,
            path=metadata.path_display,
            shared_link=None,
            raw_link=None
        )
        if self.create_shared_link:
            # create a shared link
            shared_link = dbx.sharing_create_shared_link(target_file_name)
            res['shared_link'] = shared_link.url
            res['raw_link'] = self._make_raw_file_url(shared_link.url)

        return res



import copy
from abc import abstractmethod
from collections import namedtuple
from datetime import datetime
from typing import Any, Callable, Optional

from ..models import TaskSet, PushModel
from ..selector import PayloadSelector
from ..utils import Loggable, Singleton, parse_duration_literal, DurationLiteral, auto_str, interruptible_sleep, \
    StopCycleError, is_iterable_but_no_str
from ..validator import Validator

"""
Contains directive information for engines on how to proceed on failure cases.
abort -> Abort the pull
wait_for -> how long the pulling should be interrupted after the interrupt retry the pull
retry_cnt -> Just a context information.
"""
RetryDirective = namedtuple("RetryDirective", ["abort", "wait_for", "retry_cnt"])


class NotSupportedError(Exception):
    pass


@auto_str(__repr__=True)
class Engine (Loggable):
    """
    An engine is responsible for launching a defined set of tasks and the communication between pulls and and their
    associated pushes.

    A call to to an engine's `run(...)` method will block the calling thread until the engine decides the job is done
    (normally an external SIGTERM occurs)
    """
    def __init__(self):
        pass

    def _sleep(self, sleep_time):
        def callback():
            if self.stopped.is_set():
                raise StopCycleError()
        # We couldn't just easily call time.sleep(self, retry_wait), cause we cannot react on the stopping
        # signal. So have to manually sleep in 0.5 second steps and check for stopping signal
        interruptible_sleep(sleep_time, callback, interval=0.5)

    def run(self, tasks: TaskSet):
        return self._run(tasks)

    @abstractmethod
    def _run(self, tasks: TaskSet):
        raise NotImplementedError()

    def stop(self):
        self._stop()

    @abstractmethod
    def _stop(self):
        raise NotImplementedError()


@auto_str(__repr__=True)
class RetryHandler:
    """
    A retry handler comes into play when a pull component exits unexpectedly whether by breaking the actual loop
    or by error. The `RetryHandler` decides how to proceed further (wait and retry the pull or more sophisticated
    logic).
    """
    @abstractmethod
    def handle_error(self) -> RetryDirective:
        raise NotImplementedError()


class NoRetryHandler(RetryHandler):
    """Will instruct the engine to abort and not to retry the pull."""
    def handle_error(self) -> RetryDirective:
        return RetryDirective(abort=True, wait_for=0, retry_cnt=1)


class SimpleRetryHandler(RetryHandler):
    """Simply instructs the engine to wait for the given amount of time after an error."""
    def __init__(self, retry_wait: DurationLiteral = 60, **kwargs):
        self.retry_wait = parse_duration_literal(retry_wait)
        self.retry_count = 0

    def _incr_retry(self):
        self.retry_count += 1

    def handle_error(self) -> RetryDirective:
        self._incr_retry()
        return RetryDirective(abort=False, wait_for=self.retry_wait, retry_cnt=self.retry_count)


class LimitedRetryHandler(SimpleRetryHandler):
    """Instructs the engine to wait for the given amout of time after an error. If the specified `max_retries` is hit
    the engine is instructed to abort the `pull`."""
    def __init__(self, max_retries: Optional[int] = 3, **kwargs):
        super().__init__(**kwargs)
        self.max_retries = Validator.cast_or_none(int, max_retries)

    def _eval_abort(self, retry_count: int) -> bool:
        if self.max_retries is None or self.max_retries < 0:
            return False
        return retry_count > self.max_retries

    def handle_error(self):
        super().handle_error()
        abort = self._eval_abort(self.retry_count)
        return RetryDirective(abort=abort, wait_for=self.retry_wait, retry_cnt=self.retry_count)


class AdvancedRetryHandler(LimitedRetryHandler):
    """Works like the `LimitedRetryHandler` but will reset the retry count when a given amount of time between the
    current failure and the previous failure has passed."""
    def __init__(self, reset_retry_threshold: DurationLiteral = 60, **kwargs):
        super().__init__(**kwargs)
        # Reset retry_count after x seconds of successful running
        self.reset_retry_threshold = parse_duration_literal(reset_retry_threshold)
        self.last_error = None

    def handle_error(self) -> RetryDirective:
        if self.last_error is None:
            # Initial value -> no retries so far, the next is 1
            self.retry_count = 0
        elif (datetime.now() - self.last_error).total_seconds() > self.reset_retry_threshold:
            # Reset retry count because threshold has reached, next try is 1
            self.retry_count = 0

        self.last_error = datetime.now()
        directive = super().handle_error()

        return RetryDirective(abort=directive.abort, wait_for=self.retry_wait, retry_cnt=self.retry_count)


class PushExecutor(Loggable, Singleton):
    """
    Given a payload and the push this helper actually "executes" the push by passing the payload after the (optional)
    selector did some magic to it.

    Examples:

        >>> payload = dict(a="This is the payload", b="another ignored key by selector")
        >>> from pnp.plugins.push.simple import Nop
        >>> push_instance = Nop(name='doctest')
        >>> push = PushModel(instance=push_instance, selector='data.a', deps=[], unwrap=False)
        >>> PushExecutor().execute("doctest", payload, push)
        >>> push_instance.last_payload
        'This is the payload'

    """

    def _execute_internal(self, id: str, payload: Any, push: PushModel, result_callback: Callable = None) -> None:
        self.logger.debug("[{id}] Selector: Applying '{push.selector}' to '{payload}'".format(**locals()))
        # If selector is None -> returns a dot accessable dictionary if applicable
        # If selector is not None -> returns the evaluated expression (as a dot accessable dictionary
        # if applicable
        payload = PayloadSelector.instance.eval_selector(push.selector, copy.deepcopy(payload))

        # Only make the push if the selector wasn't evaluated to suppress the push
        if payload is not PayloadSelector.instance.SuppressLiteral:
            self.logger.debug("[{id}] Emitting '{payload}' to push '{push.instance}'".format(**locals()))
            push_result = push.instance.push(payload=payload)

            # Trigger any dependent pushes
            for dependency in push.deps:
                if result_callback:
                    # Delegate work back to the engine
                    self.logger.debug("[{id}] Dependency callback is given. Delegating work back to engine".format(
                        **locals()
                    ))
                    result_callback(push_result, dependency)
                else:
                    # No delegation callback defined: The executor will recursively call itself
                    self.logger.debug("[{id}] No callback is given. Recursively process dependencies".format(
                        **locals()
                    ))
                    self.execute(id, push_result, dependency)
        else:
            self.logger.debug("[{id}] Selector evaluated to suppress literal. Skipping the push".format(**locals()))

    def execute(self, id: str, payload: Any, push: PushModel, result_callback: Callable = None) -> None:
        """
        Executes the given push by passing the specified payload.
        In concurrent environments there might be multiple executions in parallel. You may specify an `id` argument
        to identify related execution steps in the logs.
        Use the `result_callback` when the engine can take care of dependent pushes as well. The result and a dependent
        push will be passed via the callback. If the callback is not specified the PushExecute will execute them in
        a recursive manner.

        Args:
            id: ID to identify related execution steps in the logs (makes sense in concurrent environments).
            payload: The payload to pass to the push.
            push: The push instance that has to process the payload.
            result_callback: See explanation above.

        Returns:
            None.
        """
        Validator.is_instance(PushModel, push=push)

        if result_callback and not callable(result_callback):
            self.logger.warning("Result callback is given, but is not a callable. Callback will be ignored.")
            result_callback = None

        if push.unwrap and is_iterable_but_no_str(payload):
            length = len(payload)
            self.logger.debug("[{id}] Unwrapping payload to {length} individual items".format(**locals()))
            for item in payload:
                self._execute_internal(id, item, push, result_callback)
        else:
            self._execute_internal(id, payload, push, result_callback)



from importlib import import_module

from argresolver import EnvironmentResolver

from ..utils import Loggable, auto_str, auto_str_ignore
from ..validator import Validator


class InstallOptionalExtraError(ImportError):
    def __init__(self, extra_name):
        super().__init__("You have to install extra '{}' to use this plugin".format(extra_name))


def load_optional_module(namespace, extra):
    try:
        return import_module(namespace)
    except ImportError:
        raise InstallOptionalExtraError(extra)


class PluginMeta(type):
    def __new__(meta, name, bases, dct):
        newly = super().__new__(meta, name, bases, dct)
        # Force all __init__ of plugins to be decorated with envargs
        newly.__init__ = EnvironmentResolver(default_override=True)(newly.__init__)
        return newly


@auto_str(__repr__=True)
@auto_str_ignore(['_base_path'])
class Plugin(Loggable, metaclass=PluginMeta):
    def __init__(self, name, base_path=None, **kwargs):
        super().__init__()
        self.name = str(name)
        self._base_path = base_path and str(base_path)

    @property
    def base_path(self):
        # Base path is whether injected (most probably were the loaded config is located
        if self._base_path is None:
            # ... or the current working directory
            import os
            return os.getcwd()
        return self._base_path


class ClassNotFoundError(Exception):
    pass


class NamespaceNotFoundError(Exception):
    pass


class InvocationError(Exception):
    pass


class PluginTypeError(Exception):
    pass


def load_plugin(plugin_path, plugin_type, instantiate=True, **kwargs):
    """
    Loads a plugin by using a fully qualified identifier (<module_path>.<class_name>,
    e.g. pnp.plugins.pull.simple.Repeat).

    Args:
        plugin_path: Fully qualified path to plugin path.
        plugin_type: Base class the plugin has to extend / inherit from.
        instantiate: If True the class will be instantiated by passing the given **kwargs. Otherwise the class (not
            an object) is returned.
        **kwargs: Plugin arguments.

    Returns:
        If everything went fine an instantiated plugin is returned.
        Multiple things can go wrong:
        - Module not found (NamespaceNotFoundError)
        - Class not found (ClassNotFoundError)
        - Instantiation error (InvocationError)
        - Wrong plugin base type (PluginTypeError)
    """
    Validator.is_instance(str, plugin_path=plugin_path)
    Validator.is_instance(type, str, plugin_type=plugin_type)
    if isinstance(plugin_type, str) and plugin_type != 'callable':
        raise ValueError("When 'plugin_type' is str, the only allowed value is callable")

    k = plugin_path.rfind('.')
    if k > -1:
        # Namespace = everything before the last '.'
        namespace = plugin_path[:k]
        # Class name = everything after the last '.'
        clazz_name = plugin_path[k + 1:]
    else:
        namespace = 'builtins'
        clazz_name = plugin_path

    try:
        loaded_module = import_module(namespace)
        clazz = getattr(loaded_module, clazz_name)

        if isinstance(plugin_type, str) and plugin_type == 'callable':
            if not callable(clazz):
                raise PluginTypeError("The plugin is requested to be a callable, but it is not.")
        else:
            if not issubclass(clazz, plugin_type):
                raise PluginTypeError("The plugin is requested to inherit from '{}', but it does not."
                                      .format(plugin_type))

        return clazz(**kwargs) if instantiate else clazz
    except AttributeError:
        raise ClassNotFoundError('Class {} was not found in namespace {}'.format(clazz_name, namespace))
    except ImportError:
        raise NamespaceNotFoundError("Namespace '{}' not found".format(namespace))
    except TypeError as e:
        raise InvocationError("Invoked constructor from class '{}' in namespace '{}' failed".format(
            clazz_name, namespace), e)



from typing import Dict, List, Optional, Union

import attr
from box import Box

from .plugins import load_plugin
from .plugins.pull import PullBase
from .plugins.push import PushBase
from .plugins.udf import UserDefinedFunction

TaskName = str
TaskCfg = Union[Box, dict]
UDFCfg = Union[Box, dict]


def _validate_deps(instance, attrib, val):
    if val is None:
        return None
    if not isinstance(val, list):
        raise TypeError("Dependencies is expected to be a list of pushes")
    for exp_push in val:
        if not isinstance(exp_push, PushModel):
            raise TypeError("A dependency is expected to be a child of pnp.plugins.push.PushBase")


def _validate_pushes(instance, attrib, val):
    if not isinstance(val, list):
        raise TypeError("Pushes is expected to be a list of pushes")
    for exp_push in val:
        if not isinstance(exp_push, PushModel):
            raise TypeError("A push is expected to be a child of pnp.models.Push")


def _validate_callable(instance, attrib, val):
    if not callable(val):
        raise TypeError("The value is expected to be a callable")


@attr.s
class PullModel:
    instance = attr.ib(validator=attr.validators.instance_of(PullBase))


@attr.s
class PushModel:
    instance = attr.ib(validator=attr.validators.instance_of(PushBase))
    selector = attr.ib(validator=attr.validators.optional(attr.validators.instance_of((str, list, dict))), default=None)
    unwrap = attr.ib(converter=bool, default=False)
    deps = attr.ib(validator=_validate_deps, default=list())


@attr.s
class TaskModel:
    name = attr.ib(converter=str)
    pull = attr.ib(validator=attr.validators.instance_of(PullModel))
    pushes = attr.ib(validator=_validate_pushes)

    @classmethod
    def mk_pull(cls, task: TaskCfg, **extra) -> PullModel:
        args = {'name': '{task.name}_pull'.format(task=task), **extra, **task.pull.args}
        return PullModel(instance=load_plugin(task.pull.plugin, PullBase, **args))

    @classmethod
    def mk_push(cls, task: TaskCfg, **extra) -> List[PushModel]:
        def _many(pushlist, prefix):
            for i, push in enumerate(pushlist):
                push_name = '{prefix}_{i}'.format(**locals())
                args = {'name': push_name, **extra, **push.args}
                unwrap = getattr(push, 'unwrap', False)
                yield PushModel(
                    instance=load_plugin(push.plugin, PushBase, **args),
                    selector=push.selector,
                    unwrap=unwrap,
                    deps=list(_many(push.deps, push_name))
                )
        return list(_many(task.pushes, "{task.name}_push".format(**locals())))

    @classmethod
    def from_dict(cls, task: TaskCfg, base_path: Optional[str] = None) -> 'TaskModel':
        if not isinstance(task, Box):
            task = Box(dict(base=task)).base

        extra = {'base_path': base_path}
        return cls(
            name=task.name,
            pull=cls.mk_pull(task, **extra),
            pushes=list(cls.mk_push(task, **extra))
        )


@attr.s
class UDFModel:
    name = attr.ib(converter=str)
    callable = attr.ib(validator=_validate_callable)

    @classmethod
    def from_config(cls, dct: UDFCfg) -> 'UDFModel':
        if not isinstance(dct, Box):
            dct = Box(dct)
        udf_type = 'callable' if not hasattr(dct, 'args') else UserDefinedFunction
        instantiate = hasattr(dct, 'args')
        kwargs = dict() if not hasattr(dct, 'args') or dct.args is None else dct.args
        fun = load_plugin(dct.plugin, udf_type, instantiate, **{'name': dct.name, **kwargs})
        return cls(name=dct.name, callable=fun)


TaskSet = Dict[str, TaskModel]


def tasks_to_str(tasks: TaskSet):
    res = []
    for _, t in tasks.items():
        res.append('{task.name} {{'.format(task=t))
        res.append('\tpull = {task.pull}'.format(task=t))
        res.append('\tpushes = ['.format())
        for push in t.pushes:
            res.append('\t\t{push}'.format(push=push))
        res.append('\t]')
        res.append('}')
    return "\n".join(res)



"""Pull 'n' Push

Usage:
  pnp [(-c | --check)] [(-v | --verbose)] [--log=<log_conf>] <configuration>
  pnp (-h | --help)
  pnp --version

Options:
  -c --check        Only check configuration and do not run it.
  -v --verbose      Switches log level to debug.
  --log=<log_conf>  Specify logging configuration to load.
  -h --help         Show this screen.
  --version         Show version.

"""

import logging
import logging.config
import os

from docopt import docopt
from schema import Schema, Use, And, Or
from ruamel import yaml

from ..utils import get_first_existing_file
from ..app import Application


def setup_logging(*candidates, default_level=logging.INFO, env_key='PNP_LOG_CONF', verbose=False):
    """Setup logging configuration"""
    log_file_path = get_first_existing_file(*candidates)
    env_path = os.getenv(env_key, None)
    if env_path:
        log_file_path = env_path
    if log_file_path and os.path.exists(log_file_path):
        with open(log_file_path, 'rt') as f:
            config = yaml.safe_load(f.read())
        logging.config.dictConfig(config)
        logging.info("Logging loaded from: {}".format(log_file_path))
        if verbose:
            logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s',
                            level=logging.DEBUG if verbose else default_level)
        logging.info("Logging loaded with basic configuration")


def validate_args(args):
    validator = Schema({
        "--help": Use(bool),
        "--verbose": Use(bool),
        "--version": Use(str),
        "--check": Use(bool),
        "--log": Or(None, And(os.path.isfile, Use(os.path.abspath))),
        "<configuration>": And(os.path.isfile, Use(os.path.abspath))
    })
    return validator.validate(args)


def run(args):
    validated = validate_args(args)
    pnp_cfg_path = validated['<configuration>']
    default_log_level = os.environ.get('LOG_LEVEL', 'DEBUG')
    setup_logging(args['--log'], 'logging.yaml', os.path.join(os.path.dirname(pnp_cfg_path), 'logging.yaml'),
                  default_level=default_log_level, verbose=validated['--verbose'])
    app = Application.from_file(pnp_cfg_path)
    if not validated['--check']:
        app.start()


def main():
    arguments = docopt(__doc__, version='0.14.0')
    run(arguments)


if __name__ == '__main__':
    main()



import io
import os

from . import PushBase
from .. import load_optional_module
from ...utils import make_list, auto_str_ignore
from ...validator import Validator


@auto_str_ignore(['known_encodings'])
class FaceR(PushBase):
    """
    FaceR (short one for face recognition) tags known faces in images. Output is the image with the
    all faces tagged whether with the known name or an `unknown_label`. Default for unknown ones is 'Unknown'.

    Known faces can be ingested either by a directory of known faces (`known_faces_dir`) or by mapping of `known_faces`
    (dictionary: name -> [list of face files]).

    The `payload` passed to the `push` method is expected to be a valid byte array that represents an image in memory.

    Args:
        known_faces (dict): Mapping of a person's name to a list of images that contain the person's face.
        known_faces_dir (str): A directory containing images with knonwn persons (file_name -> person's name)
        unknown_label (str): Tag label of unknown faces.

    Returns:
        If the push is performed the resulting dictionary looks like
        {
            'tagged_image': <bytes of tagged image>
            'no_of_faces': 2
            'known_faces': ['obama']
        }

    Example configuration:

        name: faceR
        pull:
          plugin: pnp.plugins.pull.fs.FileSystemWatcher
          args:
            path: "/tmp"
            recursive: True
            patterns: "*.jpg"
            ignore_directories: True
            case_sensitive: False
            events: [created]
            load_file: True
            mode: binary
            base64: False
        push:
          plugin: pnp.plugins.push.ml.FaceR
          args:
            known_faces_dir: /path/to/known/faces
            unknown_label: "don't know him"
    """
    EXTRA = 'faceR'

    def __init__(self, known_faces=None, known_faces_dir=None, unknown_label="Unknown", **kwargs):
        # known_faces -> mapping name -> list of files
        # known_faces_dir -> directory with known faces (filename -> name)
        # unknown_label -> Label for unknown faces
        super().__init__(**kwargs)

        # Do not break the complete module, when extra_packages are not present
        self.face_recognition = load_optional_module('face_recognition', self.EXTRA)

        Validator.one_not_none(known_faces=known_faces, known_faces_dir=known_faces_dir)
        # If both are set known_faces is the default
        if known_faces is not None:
            # Process known faces mapping
            self.known_names, self.known_encodings = self._load_from_mapping(known_faces)
        else:
            # Process known faces directory
            Validator.is_directory(known_faces_dir=known_faces_dir)
            self.known_names, self.known_encodings = self._load_from_directory(known_faces_dir)

        self.unknown_label = str(unknown_label)

    def _load_from_mapping(self, m):
        def _loop():
            for name, fps in m.items():
                for fp in make_list(fps):
                    yield name, self._load_fencodings(fp)
        t = list(_loop())
        return [name for name, _ in t], [fp for _, fp in t]

    def _load_from_directory(self, directory):
        def _loop():
            accepted = ('.tif', '.tiff', '.gif', '.jpeg', '.jpg', '.jif', '.jfif', '.png')
            for file in os.listdir(os.fsencode(directory)):
                filename = os.fsdecode(file)
                if filename.endswith(accepted):
                    # strip the extension from the filename
                    yield os.path.splitext(filename)[0], self._load_fencodings(os.path.join(directory, filename))
        t = list(_loop())
        return [name for name, _ in t], [fp for _, fp in t]

    def _load_fencodings(self, fp):
        img = self.face_recognition.load_image_file(fp)
        return self.face_recognition.face_encodings(img)[0]

    def _tag_image(self, unknown_image, tags):
        from PIL import Image, ImageDraw

        # Convert the image to a PIL-format image so that we can draw on top of it with the Pillow library
        # See http://pillow.readthedocs.io/ for more about PIL/Pillow
        pil_image = Image.fromarray(unknown_image)
        # Create a Pillow ImageDraw Draw instance to draw with
        draw = ImageDraw.Draw(pil_image)
        try:
            for name, (top, right, bottom, left) in tags:
                # Draw a box around the face using the Pillow module
                draw.rectangle(((left, top), (right, bottom)), outline=(0, 0, 255))

                # Draw a label with a name below the face
                text_width, text_height = draw.textsize(name)
                draw.rectangle(
                    ((left, bottom - text_height - 10), (right, bottom)),
                    fill=(0, 0, 255),
                    outline=(0, 0, 255)
                )
                draw.text((left + 6, bottom - text_height - 5), name, fill=(255, 255, 255, 255))

            return pil_image
        finally:
            # Remove the drawing library from memory as per the Pillow docs
            del draw

    def push(self, payload):
        envelope, payload = self.envelope_payload(payload)
        # Load unknown image and find faces
        unknown_image = self.face_recognition.load_image_file(io.BytesIO(payload))
        face_locations = self.face_recognition.face_locations(unknown_image)
        face_encodings = self.face_recognition.face_encodings(unknown_image, face_locations)

        tags = []
        # Loop through each face found in the unknown image
        for flocation, face_encoding in zip(face_locations, face_encodings):
            # See if the face is a match for the known face(s)
            matches = self.face_recognition.compare_faces(self.known_encodings, face_encoding)
            name = self.unknown_label
            # If a match was found in known_face_encodings, just use the first one.
            if True in matches:
                first_match_index = matches.index(True)
                name = self.known_names[first_match_index]
            tags.append((name, flocation))

        # Tag the face locations
        tagged_image = self._tag_image(unknown_image, tags)

        # Save the image to a bytestream
        out_io = io.BytesIO()
        try:
            tagged_image.save(out_io, format='PNG')
            return dict(
                tagged_image=out_io.getvalue(),
                no_of_faces=len(face_locations),
                known_faces=list(set([name for name, _ in tags if name != self.unknown_label]))
            )
        finally:
            del out_io



import os
import time

from . import PushBase
from ...validator import Validator


class FileDump(PushBase):
    """
    This push dumps the given `payload` to a file to the specified `directory`.
    If argument `file_name` is None, a name will be generated based on the current datetime (%Y%m%d-%H%M%S).
    If `file_name` is not passed (or None) you should pass `extension` to specify the extension of the generated
    file name.
    Argument `binary_mode` controls whether the dump is binary (mode=wb) or text (mode=w).

    Args:
        directory (str): The target directory to store the dumps.
        file_name (str or None): The name of the file to dump. If set to None a file name will be automatically
            generated. You can specify the file_name via the envelope, too.
            Envelope will override __init__ file name.
        extension (str): The extension to use when the file name is automatically generated. Can be overridden by
            envelope.
        binary_mode (bool): If set to True the file will be written in binary mode ('wb'); otherwise in text mode ('w').

    Returns:

        When the push is performed it returns the name of the created file.

    Example configuration:

        name: file_dump
        pull:
          plugin: pnp.plugins.pull.simple.Repeat
          args:
            repeat: "Hello World"
        push:
          plugin: pnp.plugins.push.fs.FileDump
          args:
            directory: "/tmp"
            file_name: null  # Auto-generated file (timestamp)
            extension: ".txt"  # Extension of auto-generated file
            binary_mode: False  # text mode

    Examples:

        >>> import tempfile
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     # Automatic generated file name with extension '.dump'
        ...     dut = FileDump(name='doctest', directory=tmpdir, binary_mode=False)
        ...     created_file = dut.push("I am the content")  #  e.g. 20180616-123159.dump
        ...     with open(created_file, 'r') as fs:
        ...         assert fs.read() == "I am the content"

    """
    def __init__(self, directory='.', file_name=None, extension='.dump', binary_mode=True, **kwargs):
        super().__init__(**kwargs)
        self.directory = directory
        Validator.is_directory(directory=self.directory)
        self.extension = self._parse_extension(extension)
        self.binary_mode = bool(binary_mode)
        self.file_name = self._parse_file_name(file_name)

    def _parse_file_name(self, value):
        return Validator.cast_or_none(str, value)

    def _parse_extension(self, value):
        value = str(value)
        if not value.startswith('.'):
            return'.' + value
        return value

    def push(self, payload):
        envelope, payload = self.envelope_payload(payload)
        file_name = self._parse_envelope_value('file_name', envelope)
        extension = self._parse_envelope_value('extension', envelope)

        if file_name is None:
            file_name = time.strftime("%Y%m%d-%H%M%S")
        file_path = os.path.join(self.directory, file_name + extension)
        with open(file_path, 'wb' if self.binary_mode else 'w') as fs:
            fs.write(str(payload) if not self.binary_mode else payload)

        return file_path



from . import UserDefinedFunction


class Counter(UserDefinedFunction):
    """
    Example:

        >>> dut = Counter(name='doctest')
        >>> dut(), dut(), dut()
        (0, 1, 2)
        >>> dut = Counter(init=5, name='doctest')
        >>> dut(), dut(), dut()
        (5, 6, 7)
        >>> dut = Counter(init=-5, name='doctest')
        >>> dut(), dut(), dut()
        (0, 1, 2)
        >>> dut = Counter(init="5", name='doctest')
        >>> dut(), dut(), dut()
        (5, 6, 7)
    """
    def __init__(self, init=0, **kwargs):
        super().__init__(**kwargs)
        self.cnt = int(init)
        if self.cnt < 0:
            self.cnt = 0

    def action(self):
        try:
            return self.cnt
        finally:
            self.cnt += 1


class Memory(UserDefinedFunction):
    """
    Example:

        >>> dut = Memory(name='doctest', init='foo')
        >>> dut(), dut(new_memory='baz'), dut()
        ('foo', 'foo', 'baz')
    """
    _MISSING = object()

    def __init__(self, init=None, **kwargs):
        super().__init__(**kwargs)
        self.init = init

    def action(self, new_memory=_MISSING):
        try:
            return self.init
        finally:
            if new_memory is not self._MISSING:
                self.init = new_memory



import json
from abc import abstractmethod

from dictmentor import DictMentor, ext

from . import PushBase
from ...utils import try_parse_bool, try_parse_int, auto_str_ignore
from ...validator import Validator


@auto_str_ignore(['password'])
class _MQTTBase(PushBase):
    __prefix__ = 'mqtt'

    def __init__(self, host, port=1883, user=None, password=None, qos=0, **kwargs):
        super().__init__(**kwargs)
        self.host = str(host)
        self.port = int(port)
        self.user = user and str(user)
        self.password = password and str(password)
        self.qos = self._parse_qos(qos)

    def _parse_qos(self, val):
        pval = try_parse_int(val)
        if pval is None:
            self.logger.warning("QOS level of '{val}' is not int parsable. Defaulting to 0".format(**locals()))
            return 0
        if pval < 0:
            self.logger.warning("QOS level of '{val}' is invalid. Defaulting to 0".format(**locals()))
            return 0
        if pval > 2:
            self.logger.warning("QOS level of '{val}' is invalid. Defaulting to 2".format(**locals()))
            return 2
        return pval

    def _publish(self, real_payload, topic, retain=False, qos=None):
        if isinstance(real_payload, (dict, list, tuple)):
            real_payload = json.dumps(real_payload)

        auth = None
        if self.user:
            auth = dict(username=self.user, password=self.password)

        if qos is None:
            qos = self.qos

        import paho.mqtt.publish as publish
        publish.single(
            topic=topic,
            payload=real_payload,
            hostname=self.host,
            port=self.port,
            retain=retain,
            auth=auth,
            qos=qos
        )
        self.logger.debug("[{self.name}] Published message on '{topic}' @ {self.host}:{self.port} with qos={qos}. "
                          "Payload='{real_payload}'".format(**locals()))

    @abstractmethod
    def push(self, payload):
        raise NotImplementedError()


class Publish(_MQTTBase):
    """
    This push will push the given `payload` to a mqtt broker (in this case mosquitto).
    The broker is specified by `host` and `port`. In addition a topic needs to be specified were the payload
    is pushed to (e.g. home/living/thermostat).

    The `payload` will be pushed as it is. No transformation is applied. If you need to some transformations, use the
    selector.

    Returns:
        For chaining of pushes the payload is simply returned as is.

    Example configuration:

        name: mqtt
        pull:
          plugin: pnp.plugins.pull.simple.Count
        push:
          plugin: pnp.plugins.push.mqtt.MQTTPush
          args:
            host: localhost
            topic: home/devices/#
            port: 1883
    """

    def __init__(self, topic=None, retain=False, multi=False, **kwargs):
        super().__init__(**kwargs)
        self.topic = self._parse_topic(topic)
        self.retain = self._parse_retain(retain)
        self.multi = bool(multi)

    def _parse_retain(self, val):
        return try_parse_bool(val, default=False)

    def _parse_topic(self, val):
        return str(val) if val is not None else None

    @staticmethod
    def _topic_concat(t1, t2):
        if str(t1).endswith('/'):
            return t1 + t2
        return t1 + '/' + t2

    def _push_single(self, payload):
        envelope, real_payload = self.envelope_payload(payload)
        topic = self._parse_envelope_value('topic', envelope)  # Override topic via envelope
        retain = self._parse_envelope_value('retain', envelope)  # Override retain via envelope
        qos = self._parse_envelope_value('qos', envelope)  # Override qos via envelope

        if topic is None:
            raise ValueError("Topic was not defined either by the __init__ nor by the envelope")

        if not self.multi:
            self._publish(real_payload, topic, retain, qos)
        else:
            if not isinstance(real_payload, dict):
                raise TypeError("In multi mode the payload is required to be a dictionary")
            for k, v in real_payload.items():
                key_topic = self._topic_concat(topic, k)
                try:
                    self._publish(v, key_topic, retain, qos)
                except:
                    import traceback
                    ex = traceback.format_exc()
                    self.logger.error("[{self.name}] Publishing failed for message on '{key_topic}' @ "
                                      "{self.host}:{self.port} with qos={qos}. Payload='{v}'\n{ex}".format(**locals()))

    def push(self, payload):
        self._push_single(payload)
        return payload  # Payload as is. With envelope (if any).


MQTTPush = Publish


@auto_str_ignore(['configured'])
class Discovery(_MQTTBase):
    SUPPORTED_COMPONENTS = ['alarm_control_panel', 'binary_sensor', 'camera', 'cover', 'fan', 'climate', 'light',
                            'lock', 'sensor', 'switch']

    def __init__(self, discovery_prefix, component, config, object_id=None, node_id=None, **kwargs):
        super().__init__(**kwargs)
        self.discovery_prefix = str(discovery_prefix)
        Validator.is_instance(str, component=component)
        Validator.one_of(self.SUPPORTED_COMPONENTS, component=component)
        self.component = component
        self.object_id = self._parse_object_id(object_id)
        Validator.is_instance(dict, config=config)
        self._config = config
        self.node_id = self._parse_node_id(node_id)
        self.configured = {}

    @property
    def config(self):
        import copy
        return copy.copy(self._config)

    def _parse_object_id(self, val):
        return str(val)

    def _parse_node_id(self, val):
        return val and str(val)

    def _topics(self, object_id, node_id):
        node_id = "{node_id}/".format(node_id=node_id) if node_id else ""
        base_topic = "{self.discovery_prefix}/{self.component}/{node_id}{object_id}".format(**locals())
        state_topic = base_topic + "/state"
        config_topic = base_topic + "/config"
        return base_topic, config_topic, state_topic

    def _configure(self, object_id, node_id):
        configure_key = str(object_id) + str(node_id)
        if configure_key not in self.configured:
            base_topic, config_topic, state_topic = self._topics(object_id, node_id)
            mentor = DictMentor(
                ext.Variables(
                    fail_on_unset=True,
                    discovery_prefix=self.discovery_prefix,
                    component=self.component,
                    object_id=object_id,
                    node_id=node_id,
                    base_topic=base_topic,
                    config_topic=config_topic,
                    state_topic=state_topic
                )
            )
            config_augmented = mentor.augment(self.config)
            self._publish(config_augmented, config_topic, retain=True)
            self.configured[configure_key] = True

    def push(self, payload):
        envelope, real_payload = self.envelope_payload(payload)
        object_id = self._parse_envelope_value('object_id', envelope)  # Override object_id via envelope
        node_id = self._parse_envelope_value('node_id', envelope)  # Override node_id via envelope

        if object_id is None:
            raise ValueError("object_id was not defined either by the __init__ nor by the envelope")

        self._configure(object_id, node_id)
        _, _, state_topic = self._topics(object_id, node_id)
        self._publish(real_payload, state_topic, retain=True)

        return payload



from setuptools import setup, find_packages


VERSION = '0.14.0'


def readme():
    with open('README.md') as f:
        return f.read()


setup(
    name='pnp',
    version=VERSION,
    description="Pull 'n' Push",
    long_description=readme(),
    long_description_content_type='text/markdown',
    url='https://github.com/HazardDede/pnp',
    author='d.muth',
    author_email='nicetry@bitch.com',
    license='MIT',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Home Automation',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    keywords='push pull data pipeline',
    project_urls={
        'Documentation': 'https://github.com/HazardDede/pnp/blob/master/README.md',
        'Source': 'https://github.com/HazardDede/pnp/',
        'Tracker': 'https://github.com/HazardDede/pnp/issues',
    },
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=[
        'argresolver>=0.3.2',
        'asyncio>=3.4.0',
        'asyncws>=0.1',
        'attrs>=18.0.0',
        'binaryornot>=0.4.4',
        'dictmentor>=0.2.0',
        'docopt>=0.6.2',
        'psutil>=5.4.8',
        'python-box>=3.2.0',
        'pytz>=2018.4',
        'influxdb>=5.0.0',
        'paho-mqtt>=1.3.1',
        'requests>=2.18.4',
        'ruamel.yaml>=0.15.37',
        'schedule>=0.5.0',
        'schema>=0.6.7',
        'tzlocal>=1.5.0'
    ],
    extras_require={
        'dht': ['Adafruit_DHT==1.3.2'],
        'fitbit': ['fitbit>=0.3.0'],
        'fswatcher': ['watchdog>=0.8.3'],
        'faceR': ['image>=1.5.24', 'face-recognition>=1.2.2'],
        'http-server': ['Flask>=1.0.2', 'gevent>=1.3.4'],
        'gpio': ['RPi.GPIO>=0.6.5'],
        'dropbox': ['dropbox>=9.0.0', 'urllib3>=1.20'],
        'pushbullet': ['pushbullet.py>=0.10.0', 'urllib3>=1.20']
    },
    dependency_links=[
        "git+https://github.com/adafruit/Adafruit_Python_DHT.git#egg=Adafruit_DHT-1.3.2"
    ],
    python_requires='>=3.5',
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'pnp=pnp.runner.pnp:main',
        ],
    },
)



import os
from functools import partial
from ruamel import yaml

from box import Box
from schema import Schema, Use, Optional, Or, And

from .engines import Engine, RetryHandler
from .plugins import load_plugin
from .utils import make_list


class OrOverride(Or):
    """
    Validates the given data as the original Or validation directive would do, but overrides the resulting
    data.

    Example:
        >>> # Whether inbound or outbound are valid keys, but will be overridden by override
        >>> schema = Schema({OrOverride('override', 'inbound', 'outbound'): Use(str)})
        >>> schema.validate({'inbound': 'value'})  # inbound will be mapped to override
        {'override': 'value'}
        >>> schema.validate({'outbound': 'value'})  # outbound will be mapped to override
        {'override': 'value'}
        >>> OrOverride('override')  # Only override without valid directives is bad...
        Traceback (most recent call last):
        ...
        ValueError: OrMap expects the mapping value at first and then multiple validation directives

    """
    def __init__(self, *args, **kwargs):
        if len(args) < 2:
            raise ValueError("OrMap expects the mapping value at first and then multiple validation directives")
        self.override = args[0]
        super().__init__(*args[1:], **kwargs)

    def validate(self, data):
        super().validate(data)
        return self.override


push_schema = Schema({
    "plugin": Use(str),
    Optional("selector", default=None): Or(object, None),
    Optional("unwrap", default=False): bool,
    Optional("args", default={}): {
        Optional(str): object
    },
    # deps is a list of push_schemas as well, but cannot declare recursive schemas
    Optional("deps", default=list()): And(object, Use(make_list))
})

pull_schema = Schema({
    "plugin": Use(str),
    Optional("args", default={}): {
        Optional(str): object
    },
})

push_list_schema = Schema([push_schema])

task_entity_schema = Schema({
    "name": Use(str),
    OrOverride("pull", "inbound", "pull"): pull_schema,
    OrOverride("pushes", "outbound", "push"): And(Or(push_list_schema, push_schema), Use(make_list))
})

task_list_schema = Schema([task_entity_schema])

# Allow confguration of a single task
tasks = Schema(Or(And(task_entity_schema, Use(make_list)), task_list_schema))  # Single or multiple tasks are allowed

engine_schema = Engine

udfs_schema = Schema([{
    "name": Use(str),
    "plugin": Use(str),
    Optional("args"): Or({
        str: object
    }, None)
}])

tasks_settings_schema = Schema({
    Optional(Or("anchors", "anchor", "ref", "refs", "alias", "aliases")): object,
    Optional("engine", default=None): engine_schema,
    Optional("udfs", default=None): udfs_schema,
    'tasks': tasks
})

root_schema = Schema(And(
    Or(
        tasks_settings_schema,  # Tasks with additional settings and aliases (yaml only)
        tasks,  # Tasks without settings or aliases - just a list of tasks (backward compatibility)
    ),
    # In case of list only we create a tasks node - so we know it's always there
    Use(lambda x: x if 'tasks' in x else dict(tasks=x))
))

schema = Schema(root_schema)


def validate_push_deps(push):
    # Recursively validate pushes
    for dependency in push['deps']:
        validated_push = push_schema.validate(dependency)
        validated_push['deps'] = list(validate_push_deps(validated_push))
        yield validated_push


def custom_constructor(loader, node, clstype):
    args = loader.construct_mapping(node, deep=True)
    if 'type' not in args:
        raise ValueError("You have to specify a 'type' when instantiating a engine with !engine")
    clazz_name = args.pop('type')
    return load_plugin(clazz_name, clstype, **args)


def make_mentor(config_path):
    from dictmentor import DictMentor, ext
    return DictMentor(
        ext.Environment(fail_on_unset=True),
        ext.ExternalResource(base_path=os.path.dirname(config_path)),
        ext.ExternalYamlResource(base_path=os.path.dirname(config_path))
    )


def load_config(config_path):
    """Load the specified config"""
    yaml.SafeLoader.add_constructor(u"!engine", partial(custom_constructor, clstype=Engine))
    yaml.SafeLoader.add_constructor(u"!retry", partial(custom_constructor, clstype=RetryHandler))

    with open(config_path, 'r') as fp:
        cfg = yaml.safe_load(fp)
        # pnp configuration is probably of list of tasks. dictmentor needs a dict ...
        # ... let's fake it ;-)
        cfg = dict(cfg=cfg)

    mentor = make_mentor(config_path)
    # Remove the faked dictionary as root level

    validated = schema.validate(mentor.augment(cfg)['cfg'])
    for pull in validated['tasks']:
        for push in pull['pushes']:
            push['deps'] = list(validate_push_deps(push))
    b = Box(validated)

    return (
        b.udfs if 'udfs' in b else None,
        b.engine if 'engine' in b else None,
        b.tasks
    )
